// See license.txt for copyright information and redistribution information

#include <stdlib.h>
#include <math.h>

#include "SDL.h"
#include "SDL_audio.h"
#include "SFont.h"

#include "globals.h"
#include "sdl_draw_core.h"
#include "sdl_gui_core.h"
#include "particles.h"

#define VIDEOFRQ            60.0            // frequency in Hz of timer eventsddd
#define AIFRQ               15.0
#define PLAYERFRQ           45.0
#define PARTICLEFRQ         35.0            // particle "AI", life decays, & kinematics updates

#define PLAYER_GAMEOBJECT_FILE  "player.go"
#define SHAKE_GAMEOBJECT_FILE   "master_shake.go"
#define LEVEL01_MAP             "level_01.tilemap"

#define MAX_BULLETS             200              // per Game Object

#define BULLET_GFX_FILE         "bullet.bmp"
#define BULLET_LIFE             1.0             // in seconds; random factor applied
#define GOO_LIFE                6.0
#define BULLET_DMG              50
#define LASER_DMG               190
#define GOO_GFX_FILE            "shake_goo.bmp"
#define GOO_DMG                 50
#define MAX_GOO_SPEED           2.0

#define GAMEOVER_GFX_FILE       "game_over.bmp"
#define TITLE_GFX_FILE          "title.bmp"
#define HELP_GFX_FILE           "help.bmp"
#define SCORE_GFX_FILE          "scores.bmp"
#define WIN_GFX_FILE            "win.bmp"
#define GAME_FONT_FILE          "font2.bmp"

#define SHAKE_HIT_SOUND         "wabble.wav"
#define PLAYER_HIT_SOUND        "wobble.wav"
#define GOO_HIT_SOUND           "goo_hit.wav"
#define PLAYER_SHOOT_SOUND      "shoot.wav"
#define INTRO_SOUND             "intro.wav"
#define SHAKE_SHOOT_SOUND       "shake_shoot.wav"
#define PLAYER_JUMP_SOUND       "jump.wav"

#define WALK_DECAY              0.80
#define JUMP_DECAY              0.90
#define JUMP_FALL_RATE          0.20
#define MAX_FALL_SPEED          3.0

#define BLOCKS_W                20
#define BLOCKS_H                15
#define MAX_TILE_TYPES          60              // enough to fill a 320 x 192 pixel tile set image

#define INITLASERFRQ            0.2



enum { video_update, AI_update, player_update, key_repeat, particle_update, laser_eyes, no_event };



// basic game object definition for player and enemies
struct game_object {
    SDL_Rect loc;
    SDL_Rect anim_rect;
    int facing;
    float x, y;
    float vx, vy;
    int walk_flag;
    int jump_flag;
    int firing_flag;
    int active_bullets;
    float bullet_speed;
    struct particle *bullets;
    struct particle *last_bullet;
    int hp;
    int max_hp;
    int shield;
    int max_shield;
    float walk_speed;
    float jump_speed;
    int frame;
    float anim_subframe;
    SDL_Surface *gfx;
};



struct go_ai {
    float aim_x;
    float aim_y;
};



// level map - single screen ATM - from text file
struct tilemap {
    int w, h;
    int block_id[ BLOCKS_W * BLOCKS_H ];
    SDL_Rect blits[ MAX_TILE_TYPES ];
    SDL_Surface *gfx;
};



struct beam_weapon {
    float x, y;  // target position
    int status;  // firing or not
    float track_speed;  // damages GO's moving slower than this
};



// SDL Audio data
#define NUM_SOUNDS 4
struct sample {
    Uint8 *data;
    Uint32 dpos;
    Uint32 dlen;
} sounds[ NUM_SOUNDS ];



// timer callbacks
Uint32 EVNT_video_update( Uint32 interval, void *param );
Uint32 EVNT_keyrepeat( Uint32 interval, void *param );
Uint32 EVNT_player( Uint32 interval, void *param );
Uint32 EVNT_AI( Uint32 interval, void *param );
Uint32 EVNT_particles( Uint32 interval, void *param );
Uint32 EVNT_eyelasers( Uint32 interval, void *param );

// audio mixer callback
extern void mix_audio( void *unused, Uint8 *stream, int len );

// play a sound directly from disk file
void play_sound( char *file );

// main thread update functions
void update_AI();
void update_player();
void draw_debug( SDL_Surface *screen );

// simple utility functions
float clamp( float i, float max );
float sign( float i );

// game-specific utilities and GO functions
int load_game_object( const char *filename, struct game_object *go );
void decay_bullets( struct game_object *go );
void add_bullet( struct game_object *go );
struct particle *delete_bullet( struct game_object *go, struct particle *prev_p );
void move_bullets( struct game_object *go );
void draw_lifemeters();
void draw_score();
void draw_eyelasers();
void good_swarm_AI( struct particle *p );
void damage_go( struct game_object *go, int damage );

// collision routines
void go_tile_collision( struct game_object *go, struct tilemap *tm );
void particle_go_collision( struct game_object *go, struct game_object *go2 );
void bullet_goo_collision();

// tilemap routines
int load_tilemap( const char *filename, struct tilemap *tm );
void draw_tilemap( SDL_Surface *screen, const struct tilemap *tm );

// game state routines
void game_over();
int title_screen();
void score_screen();
void help_screen();
int play_state();
void empty_scores();
void you_win();



// global variables
struct particle crosshair = { NULL, 0, XREZ / 2, YREZ / 2, 0, 0, { 0, 0, 32, 32 },
    { 0, 0, 32, 32 }, sprite, NONE, 0, no_AI, IMMORTAL, NULL };

// game objects
struct game_object player;
struct game_object shake;
struct go_ai shake_ai;
struct beam_weapon laser;

// level tilemap
struct tilemap level1;

// mouse location - updated on SDL event MOUSEMOTION
int mouse_x, mouse_y;

// video display
SDL_Surface *screen;

// system bitmap font
SDL_Surface *Font;

// buffer to hold sprintf()'s for PutString()
char buffer[ 256 ];

// urm
long int score;



int main( int argc, char **argv )
{
    // take care of little vs. big endian issues
    if( BPPIXEL == 32 ) {
        #if SDL_BYTEORDER == SDL_BIG_ENDIAN
            rmask = 0xff000000;
            gmask = 0x00ff0000;
            bmask = 0x0000ff00;
            amask = 0x000000ff;
        #else
            rmask = 0x000000ff;
            gmask = 0x0000ff00;
            bmask = 0x00ff0000;
            amask = 0xff000000;
        #endif
    }
    else if( BPPIXEL == 24 ) {
        #if SDL_BYTEORDER == SDL_BIG_ENDIAN
            rmask = 0xff000000;
            gmask = 0x00ff0000;
            bmask = 0x0000ff00;
            amask = 0x00000000;
        #else
            rmask = 0x000000ff;
            gmask = 0x0000ff00;
            bmask = 0x00ff0000;
            amask = 0x00000000;
        #endif
    }

    // initialize SDL video
    if( SDL_Init( SDL_INIT_TIMER | SDL_INIT_VIDEO ) < 0 ) {
        printf( "Unable to init SDL: %s\n", SDL_GetError() );
        return 1;
    }

    // title bar and icon names
    SDL_WM_SetCaption( "Frogger -X- Master Shake", "Detox M-LD-1" );

    // SDL clean up before exit
    atexit( SDL_Quit );

    // create a new window
    screen = SDL_SetVideoMode( XREZ, YREZ, BPPIXEL, SDL_HWSURFACE|SDL_DOUBLEBUF );
    if ( !screen ) {
        printf( "Unable to open window/screen display: %s\n", SDL_GetError() );
        return 2;
    }

    // Initialize audio system
    SDL_AudioSpec fmt;

    /* Set 16-bit stereo audio at 22Khz */
    fmt.freq = 22050;
    fmt.format = AUDIO_S16;
    fmt.channels = 2;
    fmt.samples = 512;        /* A good value for games */
    fmt.callback = mix_audio;
    fmt.userdata = NULL;

    /* Open the audio device and start playing sound! */
    if ( SDL_OpenAudio( &fmt, NULL ) < 0 ) {
        printf( "Unable to open audio: %s\n", SDL_GetError() );
        return 3;
    }

    // unpause all audio
    SDL_PauseAudio( 0 );

    // game start!!!
    play_sound( INTRO_SOUND );



    ///--------------------------------------------------------------------------------------
    ///GAME LOGIC

    int playstate_code = 0;
    while( playstate_code != -3 ) {

        playstate_code = title_screen();

        if( playstate_code == -1 ) {
            printf( "Player wins! score = %d\n", score );
            you_win();
        }

        else if( playstate_code == -2 ) {
            printf( "Player lost. score = %d\n", score );
            game_over();
        }
    }

    printf( "Exiting main() cleanly\n" );
    SDL_CloseAudio();
    return 0;
}



int play_state()
{
    // Load a font and get it ready for use (SFONT requried)
	Font = (SDL_Surface *) SDL_LoadBMP( GAME_FONT_FILE );
	if( ! Font ) {
	    printf( "failed to load font BMP\n" );
	    return 16;
	}
	InitFont( Font );

    // add timer callback functions
    SDL_TimerID video_timer_id = SDL_AddTimer( (Uint32)( 1000.0 / VIDEOFRQ ), EVNT_video_update, NULL );
	if( video_timer_id == NULL ) {
	    printf( "Error adding video timer function.\n" );
	    return 5;
	}

	SDL_TimerID AI_timer_id = SDL_AddTimer( (Uint32)( 1000.0 / AIFRQ ), EVNT_AI, NULL );
	if( AI_timer_id == NULL ) {
	    printf( "Error adding AI timer function.\n" );
	    return 6;
	}

	SDL_TimerID player_timer_id = SDL_AddTimer( (Uint32)( 1000.0 / PLAYERFRQ ), EVNT_player, NULL );
	if( player_timer_id == NULL ) {
	    printf( "Error adding player timer function.\n" );
	    return 7;
	}

	SDL_TimerID particle_timer_id = SDL_AddTimer( (Uint32)( 1000.0 / PARTICLEFRQ ), EVNT_particles, NULL );
	if( particle_timer_id == NULL ) {
	    printf( "Error adding particle timer function.\n" );
	    return 12;
	}

	SDL_TimerID shakelaser_timer_id = SDL_AddTimer( (Uint32)( 1000.0 / INITLASERFRQ ), EVNT_eyelasers, NULL );
	if( shakelaser_timer_id == NULL ) {
	    printf( "Error adding eye laser timer function.\n" );
	    return 17;
	}

    // load mouse crosshair sprite
    particle_loadgfx( &crosshair, "crosshair.bmp" );

    // load player game object data and graphics
    if( load_game_object( PLAYER_GAMEOBJECT_FILE, &player ) != 0 ) {
        printf( "failed to load player data\n" );
        return 11;
    }

    // player initial game position
    player.x = XREZ / 2 - 32;
    player.y = 32 * 12;
    player.loc.x = player.x;
    player.loc.y = player.y;

    // initialize player bullet particles
    player.bullets = new_particle( NULL, sprite, no_AI, DO_DAMAGE, BULLET_DMG, BULLET_LIFE );
    set_position( player.bullets, player.x, player.y );
    particle_loadgfx( player.bullets, BULLET_GFX_FILE );
    player.last_bullet = player.bullets;

    // Shake load and init
    if( load_game_object( SHAKE_GAMEOBJECT_FILE, &shake ) != 0 ) {
        printf( "failed to load shake data\n" );
        return 14;
    }

    // Shake initial game position
    shake.x = XREZ / 2;
    shake.y = 64;
    shake.loc.x = shake.x;
    shake.loc.y = shake.y;

    // eye beam inits
    laser.status = FALSE;
    laser.x = player.x;
    laser.y = player.y;
    laser.track_speed = 1.5;

    // Shake Goo setup
    shake.bullets = new_particle( NULL, sprite, no_AI, DO_DAMAGE, GOO_DMG, GOO_LIFE );
    set_position( shake.bullets, shake.x, shake.y );
    particle_loadgfx( shake.bullets, GOO_GFX_FILE );
    shake.last_bullet = shake.bullets;

    // load level 1
    if( load_tilemap( LEVEL01_MAP, &level1 ) != 0 ) {
        printf( "failed to load level 1 tilemap.\n" );
        return 13;
    }

    // reset score
	score = 0;



    // main loop variables
    int done = FALSE;
    SDL_Event event;
    char screenshot_filename[ 80 ];
    int screenshot_number = 1;

    printf( "entering main loop\n" );

    ///----------------------------------------------------------------------------------------------------
    while ( !done ) {

        // check game over conditions
        if( player.hp < 1 || shake.hp < 1 ) done = TRUE;

        // wait for an SDL event to process
        SDL_WaitEvent( &event );

        switch( event.type ) {

            // exit if the window is closed
            case SDL_QUIT:
                printf( "SDL EVENT: window close button pressed.  Exiting...\n" );
                done = TRUE;
                break;

            // special purpose user events - timer pushed events, etc.
            case SDL_USEREVENT:

                // key repeat event (from timer)
                switch( event.user.code ) {

                    case video_update:
                        draw_tilemap( screen, &level1 );
                        SDL_BlitSurface( player.gfx, &player.anim_rect, screen, &player.loc );
                        SDL_BlitSurface( shake.gfx, &shake.anim_rect, screen, &shake.loc );
                        particle_draw( screen, player.bullets->next );
                        particle_draw( screen, shake.bullets->next );
                        draw_eyelasers();
                        particle_draw( screen, &crosshair );
                        draw_debug( screen );
                        draw_lifemeters();
                        draw_score();
                        SDL_Flip( screen );
                        break;

                    case AI_update:
                        update_AI();
                        break;

                    case player_update:
                        update_player();
                        break;

                    case particle_update:
                        decay_bullets( &player );
                        move_bullets( &player );
                        decay_bullets( &shake );
                        move_bullets( &shake );
                        bullet_goo_collision();
                        particle_go_collision( &player, &shake );
                        particle_go_collision( &shake, &player );
                        break;

                    case laser_eyes:
                        if( laser.status == TRUE ) {
                            if( ( fabs( player.vx ) + fabs( player.vy ) ) < laser.track_speed ) {
                                damage_go( &player, LASER_DMG );
                            }
                        }
                        break;

                    default:
                        break;

                }  // user event type switch

                break;  // user event entry in SDL event type switch


            // check for keypresses
            case SDL_KEYDOWN:

                switch( event.key.keysym.sym ) {

                    case SDLK_w:
                        if( player.jump_flag == FALSE ) {
                            player.vy = -player.jump_speed;
                            player.jump_flag = TRUE;
                            if( player.facing == -1 ) player.frame = 7;
                            else player.frame = 3;
                            play_sound( PLAYER_JUMP_SOUND );
                        }
                        break;

                    case SDLK_a:
                        player.vx = -player.walk_speed;
                        player.frame = 0;
                        player.walk_flag = TRUE;
                        player.facing = -1;
                        break;

                    case SDLK_d:
                        player.vx = player.walk_speed;
                        player.frame = 4;
                        player.walk_flag = TRUE;
                        player.facing = 1;
                        break;

                    case SDLK_LSHIFT:
                        player.firing_flag = TRUE;
                        break;

                    case SDLK_SPACE:
                        if( player.jump_flag == FALSE ) {
                            player.vy = -player.jump_speed;
                            player.jump_flag = TRUE;
                            play_sound( PLAYER_JUMP_SOUND );
                        }
                        break;

                    case SDLK_F12:
                        sprintf( screenshot_filename, "MLD1_screenshot_%d.bmp", screenshot_number );
                        printf( "..SCREENSHOT.. #%d\n", screenshot_number );
                        screenshot_number++;
                        if( SDL_SaveBMP( screen, screenshot_filename ) != 0 ) {
                            printf( "SDL failed to save screen shot.\n" );
                        }
                        break;

                    case SDLK_ESCAPE:
                        printf( "[ESC] KEY pressed.  Exiting...\n" );
                        done = TRUE;
                        play_sound( PLAYER_JUMP_SOUND );
                        break;

                    default:
                        break;

                } // switch on key event

                break; //SDL_KEYDOWN switch entry

            // Keyup event, probably time to cancel any action key repeat events
            case SDL_KEYUP:

                switch( event.key.keysym.sym  ) {

                    case SDLK_a:
                        player.walk_flag = FALSE;
                        player.facing = -1;
                        break;

                    case SDLK_d:
                        player.walk_flag = FALSE;
                        player.facing = 1;
                        break;

                    case SDLK_LSHIFT:
                        player.firing_flag = FALSE;
                        break;

                    default:
                        break;

                } // KEYUP switch

                break;  // SDL_KEYUP switch entry

            // Highlight gui-buttons on mouse button down
            case SDL_MOUSEBUTTONDOWN:

                if( event.button.button == SDL_BUTTON_LEFT ) {

                    // continuous weapon firing
                    player.firing_flag = TRUE;

                } // LEFT button down if statement

                break;  // SDL_MOUSEBUTTONDOWN switch entry

            // if we are in draw mode, check for a mouse movement event
            case SDL_MOUSEMOTION:

                // update global variables
                mouse_x = event.motion.x;
                mouse_y = event.motion.y;

                // update crosshair position
                set_position( &crosshair, mouse_x, mouse_y );

                break;  // SDL_MOUSEMOTION switch entry

            // perform button action on mousebutton up
            case SDL_MOUSEBUTTONUP:

                if( event.button.button == SDL_BUTTON_LEFT  ) {

                    // stop firing weapon
                    player.firing_flag = FALSE;

                }  // LEFT MOUSE BUTTON if statement

                break; // SDL_MOUSEBUTTONUP switch entry

            // SDL event type, no match found
            default:
                break;

        } // end switch on SDL event type

    } // while !done

    // remove the timer callbacks
    SDL_RemoveTimer( video_timer_id );
    SDL_RemoveTimer( particle_timer_id );
    SDL_RemoveTimer( AI_timer_id );
    SDL_RemoveTimer( player_timer_id );

    if( shake.hp < 1 ) return -1;
    if( player.hp < 1 ) return -2;

    printf( "game state DONE & no end-game condition found. Player Exit\n" );
    return -3;  // signal an exit from main game logic while() loop
}



// ---------------------------------------------------------------------------------------------------
// timer callback function for video update - signals a user event
Uint32 EVNT_video_update( Uint32 interval, void *param )
{
    //int data = FALSE;
    SDL_Event new_event;
    SDL_UserEvent user_event;


    // if we were passed a valid pointer grab it's contents as an integer value
    //if( param ) data = *( (int *) param );

    // setup the userevent struct and event struct
    user_event.type = SDL_USEREVENT;
    user_event.code = video_update;
    user_event.data1 = NULL;
    user_event.data2 = NULL;

    // setup the generic SDL event as a user event
    new_event.type = SDL_USEREVENT;
    new_event.user = user_event;

    // add user event to SDL event stack
    SDL_PushEvent( &new_event );

    // return the timer interval to call this SDL event again at same timer delta
    return interval;
}



// ---------------------------------------------------------------------------------------------------
// timer callback function
Uint32 EVNT_player( Uint32 interval, void *param )
{
    //int data = FALSE;
    SDL_Event new_event;
    SDL_UserEvent user_event;


    // if we were passed a valid pointer grab it's contents as an integer value
    //if( param ) data = *( (int *) param );

    // setup the userevent struct and event struct
    user_event.type = SDL_USEREVENT;
    user_event.code = player_update;
    user_event.data1 = NULL;
    user_event.data2 = NULL;

    // setup the generic SDL event as a user event
    new_event.type = SDL_USEREVENT;
    new_event.user = user_event;

    // add user event to SDL event stack
    SDL_PushEvent( &new_event );

    // return the timer interval to call this SDL event again at same timer delta
    return interval;
}



// ---------------------------------------------------------------------------------------------------
// timer callback function
Uint32 EVNT_eyelasers( Uint32 interval, void *param )
{
    //int data = FALSE;
    SDL_Event new_event;
    SDL_UserEvent user_event;


    // if we were passed a valid pointer grab it's contents as an integer value
    //if( param ) data = *( (int *) param );

    // setup the userevent struct and event struct
    user_event.type = SDL_USEREVENT;
    user_event.code = laser_eyes;
    user_event.data1 = NULL;
    user_event.data2 = NULL;

    // setup the generic SDL event as a user event
    new_event.type = SDL_USEREVENT;
    new_event.user = user_event;

    // add user event to SDL event stack
    SDL_PushEvent( &new_event );

    laser.x = player.x;
    laser.y = player.y;

    // return a new timer delta with a little random sauce
    if( laser.status == TRUE ) {
        laser.status = FALSE;
        return (Uint32)( 0.4 * ( 1000.0 / INITLASERFRQ ) + 3.0 * ( rand() % 100 / 100.0 ) );
    }
    else {
        laser.status = TRUE;
        return (Uint32)( 1000.0 / 1.0 );
    }
}



// ---------------------------------------------------------------------------------------------------
// timer callback function
Uint32 EVNT_particles( Uint32 interval, void *param )
{
    //int data = FALSE;
    SDL_Event new_event;
    SDL_UserEvent user_event;


    // if we were passed a valid pointer grab it's contents as an integer value
    //if( param ) data = *( (int *) param );

    // setup the userevent struct and event struct
    user_event.type = SDL_USEREVENT;
    user_event.code = particle_update;
    user_event.data1 = NULL;
    user_event.data2 = NULL;

    // setup the generic SDL event as a user event
    new_event.type = SDL_USEREVENT;
    new_event.user = user_event;

    // add user event to SDL event stack
    SDL_PushEvent( &new_event );

    // return the timer interval to call this SDL event again at same timer delta
    return interval;
}



// ---------------------------------------------------------------------------------------------------
// timer callback function
Uint32 EVNT_AI( Uint32 interval, void *param )
{
    //int data = FALSE;
    SDL_Event new_event;
    SDL_UserEvent user_event;


    // if we were passed a valid pointer grab it's contents as an integer value
    //if( param ) data = *( (int *) param );

    // setup the userevent struct and event struct
    user_event.type = SDL_USEREVENT;
    user_event.code = AI_update;
    user_event.data1 = NULL;
    user_event.data2 = NULL;

    // setup the generic SDL event as a user event
    new_event.type = SDL_USEREVENT;
    new_event.user = user_event;

    // add user event to SDL event stack
    SDL_PushEvent( &new_event );

    // return the timer interval to call this SDL event again at same timer delta
    return interval;
}



void goo_swarm_AI( struct particle *p )
{
    float i, j;

    while( p != NULL ) {

        i = player.x - p->loc.x;
        j = player.y - p->loc.y;

        p->vx += i / XREZ;
        p->vy += j / YREZ;

        if( p->vx > MAX_GOO_SPEED ) p->vx = MAX_GOO_SPEED;
        if( p->vy > MAX_GOO_SPEED ) p->vy = MAX_GOO_SPEED;

        p = p->next;
    }
}



void update_AI()
{
    // master shake movement state
        // if player_y < xxx for first time engage turn to face player mode
        // if player dist < xxx && player_y < shake_y evade


    // loop through goo shots and turn them towards player
    goo_swarm_AI( shake.bullets->next );


    // hp and shield clamps
    if( shake.hp < 0 ) {
        shake.hp = 0;
    }

    if( shake.shield < 0 ) {
        shake.shield = 0;
    }

    // update range weapon aim for goo blaster
    float i = player.x - shake.x;
    float j = player.y - shake.y;
    if( j > 2.0 ) j = 2.0; // cannot shoot downwards
    float denom = 1.0 / sqrt( i * i + j * j );
    shake_ai.aim_x = i * denom;
    shake_ai.aim_y = j * denom;

    add_bullet( &shake );
    play_sound( SHAKE_SHOOT_SOUND );


        // if player dist short range random: hop-stun/slapfight
        // if player dist med range & have LOS spew some shake goo
        // if player dist med or long & have no LOS roll out shake ball
        // if player dist long range check for portal spawn availability

    // master shake portal-spawning state
        // timer from game start xxx
        // when it runs out portal spawns are available - load some gfx tiles to show
}



/// ----------------------------------------------------------------------------------------------
void update_player()
{
    // health and shield clamps
    if( player.shield < 0 ) {
        player.shield = 0;
    }

    if( player.hp < 1 ) {
        player.hp = 0;
    }

    // decay horizontal movement velocity when key is released
    if( player.walk_flag == FALSE && player.vx != 0.0 ) {

        player.vx *= WALK_DECAY;

        if( player.vx > -0.05 && player.vx < 0.05 ) {

            player.vx = 0.0;

            if( player.facing == -1 ) player.frame = 4;
            else player.frame = 0;
        }
    }

    // apply jump function to player vertical velocity while they are jumping
    if( player.jump_flag == TRUE ) {

        if( player.vy < -0.1 ) player.vy *= JUMP_DECAY;

        else {
            player.vy += JUMP_FALL_RATE;
            player.vy = clamp( player.vy, MAX_FALL_SPEED );
        }
    }

    player.x += player.vx;
    player.y += player.vy;

    if( player.x > XREZ - player.loc.w ) player.x = XREZ - player.loc.w;
    if( player.x < 0 ) player.x = 0;
    if( player.y > YREZ - player.loc.h ) player.y = YREZ - player.loc.h;
    if( player.y < 0 ) player.y = 0;

    player.loc.x = player.x;
    player.loc.y = player.y;

    // check player's new position against tilemap; adjust position & velocities as necessary
    go_tile_collision( &player, &level1 );

    // anim frame update
    if( fabs( player.vx ) > 0.05 ) {

        player.anim_subframe += fabs( 0.5 * player.vx / player.walk_speed );
        if( player.anim_subframe >= 4.0 ) player.anim_subframe = 0;

        player.frame = player.anim_subframe;
        if( player.vx < 0 ) player.frame += 4;
    }

    // calculate the offset for blitting
    player.anim_rect.x = player.frame * 32;

    // is the player firing?
    if( player.firing_flag == TRUE ) {

        // add the new particles to the bullet list
        add_bullet( &player );
        play_sound( PLAYER_SHOOT_SOUND );
    }
}



void add_bullet( struct game_object *go )
{
    struct particle *new_bullet = NULL;


    if( go->active_bullets < MAX_BULLETS ) {

        // each bullet costs a point or 2
        score -= 1;
        if( score < 0 ) score = 0;

        // copy the base particle
        new_bullet = copy_particle( go->bullets );

        // ouch! costly math and such a terrible logic hack
        float i, j, denom, ni = 0, nj = 0;
        if( go == &player ) {
            i = mouse_x - go->x;
            j = mouse_y - go->y;
            if( j > -5.0 ) j = -5.0; // cannot shoot downwards
            denom = 1.0 / sqrt( i * i + j * j );
            ni = i * denom;
            nj = j * denom;
        }

        if( go == &shake ) {
            ni = shake_ai.aim_x;
            nj = shake_ai.aim_y;
        }

        // move start location along i,j vector by about half player blit size
        new_bullet->x = go->x + ( go->loc.w / 2 ) + ni * go->loc.w;
        new_bullet->y = go->y + ( go->loc.h / 2 ) + nj * go->loc.h;

        // give it some velocity based on the mouse cursor location and distance
        set_velocity( new_bullet, go->bullet_speed * ni, go->bullet_speed * nj );

        // random lifespan
        if( go == &player ) {
            new_bullet->lifespan = ( BULLET_LIFE * PARTICLEFRQ + PARTICLEFRQ * 0.5 * ( rand() % 100 / 100.0 ) );
        }
        else if( go == &shake ) {
            new_bullet->lifespan = ( GOO_LIFE * PARTICLEFRQ + PARTICLEFRQ * 2.0 * ( rand() % 100 / 100.0 ) );
        }


        // update the book keeping pointers
        go->last_bullet->next = new_bullet;
        go->last_bullet = new_bullet;

        go->active_bullets++;
    }
}



void move_bullets( struct game_object *go )
{
    struct particle *p = go->bullets->next;
    if( p == NULL ) return;

    int h = p->loc.h;
    int w = p->loc.w;

    // not quite tuned right...
    do {
        p->x += p->vx;
        p->y += p->vy;

        if( p->x < w ) {
             p->x = w;
             p->vx = -( p->vx );
        }

        if( p->y < h ) {
            p->y = h;
            p->vy = -( p->vy );
        }

        if( p->x > XREZ - w ) {
            p->x = XREZ - w;
            p->vx = -( p->vx );
        }

        if( p->y > YREZ - h ) {
            p->y = YREZ;
            p->vy = -( p->vy );
        }

        p = p->next;

    } while( p != NULL );
}



float clamp( float i, float max )
{
    if( fabs( i ) > max ) {
        return sign( i ) * max;
    }

    else return i;
}



float sign( float i )
{
    if( i < 0 ) return -1;
    else if( i > 0 ) return 1;
    else return 0;
}



int load_tilemap( const char *filename, struct tilemap *tm )
{
    FILE *f = NULL;
    char gfx_filename[ 80 ];
    int max_id = 0;
    int id = 0;
    int n;


    f = fopen( filename, "r" );
    if( !f ) {
        printf( "Failed to open tile map file: %s\n", filename );
        return -1;
    }

    // load size info ... not really needed ...
    if( fscanf( f, "%d, %d", &tm->w, &tm->h ) != 2 ) return -2;

    // load tilemap
    for( n = 0; n < tm->w * tm->h; n++ ) {
        if( fscanf( f, "%d", &id ) != 1 ) return -3;
        tm->block_id[ n ] = id;
        if( id > max_id ) max_id = id;
    }

    // load graphic data filename
    if( fscanf( f, "%s\n", gfx_filename ) != 1 ) return -4;

    fclose( f );

    // load BMP
    tm->gfx = SDL_LoadBMP( gfx_filename );
    if( tm->gfx == NULL ) {
        printf( "Error loading level map graphics BMP.\n" );
        return -5;
    }

    SDL_SetColorKey( tm->gfx, SDL_SRCCOLORKEY|SDL_RLEACCEL, 0 );  // 0 is transparent
    tm->gfx = SDL_DisplayFormat( tm->gfx );


    double row, offset;

    // calculate blitting rectangles
    for( n = 0; n <= max_id; n++ ) {

        tm->blits[ n ].w = 32;
        tm->blits[ n ].h = 32;

        offset = modf( n / (float) tm->w, &row );

        tm->blits[ n ].x = ( n - tm->w * (int) row ) * 32;
        tm->blits[ n ].y = 32 * row;
    }

    return 0;
}



void draw_tilemap( SDL_Surface *screen, const struct tilemap *tm )
{
    int x, y;
    int id = 0;
    SDL_Rect dest;


    dest.w = 32;
    dest.h = 32;

    for( y = 0; y < tm->h; y++ ) {

        for( x = 0; x < tm->w; x++ ) {

            id = tm->block_id[ x + y * tm->w ];

            dest.x = x * 32;
            dest.y = y * 32;

            SDL_BlitSurface( tm->gfx, &tm->blits[ id ], screen, &dest );
        }
    }
}



int load_game_object( const char *filename, struct game_object *go )
{
    FILE *f = NULL;
    char gfx_filename[ 256 ];


    f = fopen( filename, "r" );
    if( !f ) {
        printf( "Failed to open game object file: %s\n", filename );
        return -1;
    }

    if( fscanf( f, "%d\n", &go->hp ) != 1 ) return -2;
    if( fscanf( f, "%d\n", &go->shield ) != 1 ) return -3;
    if( fscanf( f, "%f, %f\n", &go->walk_speed, &go->jump_speed ) != 2 ) return -4;
    if( fscanf( f, "%f\n", &go->bullet_speed ) != 1 ) return -6;
    if( fscanf( f, "%d, %d\n", &go->loc.w, &go->loc.h ) != 2 ) return -7;
    if( fscanf( f, "%s\n", gfx_filename ) != 1 ) return -5;

    fclose( f );

    // load BMP
    go->gfx = SDL_LoadBMP( gfx_filename );
    if( go->gfx == NULL ) {
        printf( "Error loading player sprite BMP.\n" );
        return -6;
    }

    // SDL rect width & heights
    //go->loc.w = 32;
    //go->loc.h = 32;
    go->anim_rect.w = go->loc.w;
    go->anim_rect.h = go->loc.h;
    go->anim_rect.y = 0;

    // set default values
    go->firing_flag = FALSE;
    go->walk_flag = FALSE;
    go->jump_flag = FALSE;
    go->frame = 0;
    go->anim_subframe = 0.0;
    go->facing = 1;
    go->active_bullets = 0;
    go->bullets = NULL;
    go->last_bullet = NULL;
    go->max_hp = go->hp;
    go->max_shield = go->shield;

    // set transparent color and RLE accelerated blitting
    SDL_SetColorKey( go->gfx, SDL_SRCCOLORKEY|SDL_RLEACCEL, 0 );
    go->gfx = SDL_DisplayFormat( go->gfx );
    printf( "Player surface loaded.\n" );

    return 0;
}



void damage_go( struct game_object *go, int damage )
{
    if( go->shield > 0 ) {
        go->shield -= damage;
    }
    else go->hp -= damage;
}



void draw_debug( SDL_Surface *screen )
{
/*    sprintf( buffer, "vx: %.3f", player.vx );
    PutString( screen, 0, 30, buffer );
    sprintf( buffer, "vy: %.3f", player.vy );
    PutString( screen, 0, 45, buffer );
    sprintf( buffer, "bullets: %d", player.active_bullets );
    PutString( screen, 0, 60, buffer );
    sprintf( buffer, "goo: %d", shake.active_bullets );
    PutString( screen, 0, 75, buffer );*/
}



// only works for 32x32 tile
// wow this is ugly
void go_tile_collision( struct game_object *go, struct tilemap *tm )
{
    // calculate tilemap blocks
    int ax0 = go->loc.x;
    int ax1 = ax0 + go->loc.w;
    int axc = ax0 + go->loc.w / 2;
    int ay0 = go->loc.y;
    int ay1 = ay0 + go->loc.h;
    int ayc = ay0 + go->loc.h / 2;
    int cl = ( ax0 / 32 ) + ( ayc / 32 ) * tm->w;
    int cr = ( ax1 / 32 ) + ( ayc / 32 ) * tm->w;
    int cu = ( axc / 32 ) + ( ay0 / 32 ) * tm->w;
    int cd = ( axc / 32 ) + ( ay1 / 32 ) * tm->w;
    int ul = ( ax0 / 32 ) + ( ay0 / 32 ) * tm->w;
    int ur = ( ax1 / 32 ) + ( ay0 / 32 ) * tm->w;
    int ll = ( ax0 / 32 ) + ( ay1 / 32 ) * tm->w;
    int lr = ( ax1 / 32 ) + ( ay1 / 32 ) * tm->w;
    if( ax1 > XREZ ) {
        ur = 1;
        lr = 1;
        cr = 1;
    }
    if( ax0 < 0 ) {
        ul = 1;
        ll = 1;
        cl = 1;
    }
    int centerright = tm->block_id[ cr ];
    int centerleft = tm->block_id[ cl ];
    int centerup = tm->block_id[ cu ];
    int centerdown = tm->block_id[ cd ];
    int upperleft = tm->block_id[ ul ];
    int upperright = tm->block_id[ ur ];
    int lowerleft = tm->block_id[ ll ];
    int lowerright = tm->block_id[ lr ];

    // waling right
    if( go->vx > 0 ) {
        if( centerright > 0 ) {
            go->x -= go->vx;
        }
    }

    // walking left
    if( go->vx < 0 ) {
        if( centerleft > 0 ) {
            go->x -= go->vx;
        }
    }


    int centerdown2 = tm->block_id[ ( ( ax0 + go->loc.w / 2 ) / 32 ) + ( ( ay1 + 1 ) / 32 ) * tm->w ];

    // jumping - look for the ground - can catch a toe on edge - check the lower corners
    if( go->vy > 0 ) {
        if( lowerleft > 0 || lowerright > 0 ) {
        //if( centerdown > 0 ) {
            go->y -= go->vy;
            go->vy = 0;
            go->jump_flag = FALSE;
        }
    }

    // falling - walked off edge - engage jump mode and start out with some -vy
    if( centerdown == 0 && go->jump_flag == FALSE ) {

        // one of the two corners has to be over open space too
        if( lowerleft == 0 || lowerright == 0 ) {

            // nudge left/right if we're on the edge
            if( lowerleft > 0 ) {
                go->x += 1;
            }

            if( lowerright > 0 ) {
                go->x -= 1;
            }

            // check next pixels down to remove jitter
            //int lowerleft2 = tm->block_id[ ( ax0 / 32 ) + ( ( ay1 + 1 ) / 32 ) * tm->w ];
            //int lowerright2 = tm->block_id[ ( ax1 / 32 ) + ( ( ay1 + 1 ) / 32 ) * tm->w ];
            //int centerdown2 = tm->block_id[ ( ( ax0 + go->loc.w / 2 ) / 32 ) + ( ( ay1 + 1 ) / 32 ) * tm->w ];
            //if( lowerleft2 == 0 && lowerright2 == 0 ) {
            if( centerdown2 == 0 ) {
                go->vy = 0.1;
                go->jump_flag = TRUE;
            }
        }
    }


    // hitting your head on tiles... checks against the upper corners
    if( go->vy < 0 ) {
        if( upperleft > 0 || upperright > 0 ) {
            go->y -= go->vy;
            go->vy = 0;
        }
    }
}


// delete's next particle in list (cannot be NULL!)
// return the new next particle in the list (can be NULL)
struct particle *delete_bullet( struct game_object *go, struct particle *prev_p )
{
    // soon-to-be next particle in list after we delete...
    struct particle *temp_next = prev_p->next->next;

    // free the next particle
    free( prev_p->next );

    // update the next pointer
    prev_p->next = temp_next;

    // if next pointer is NULL update the last_bullet pointer in the GameObject
    if( prev_p->next == NULL ) go->last_bullet = prev_p;

    go->active_bullets--;

    // return the new next in list
    return prev_p->next;
}



// skips the first in the list - first in list is a prototype to be copied
void decay_bullets( struct game_object *go )
{
    struct particle *p = go->bullets->next;
    struct particle *previous = go->bullets;
    struct particle *temp = NULL;

    // no bullets in use
    if( p == NULL ) return;

    do {
        p->lifespan--;

        if( p->lifespan < 0 ) {
            p = delete_bullet( go, previous );
        }

        else {
            // incriment down the list
            previous = p;
            p = p->next;
        }

    } while( p != NULL );
}



// skips the first in the list - first in list is a prototype to be copied
void particle_go_collision( struct game_object *target_go, struct game_object *bullet_go )
{
    int bx0, bx1, by0, by1;
    int collision;


    // Calculate GO sprite bounding box (temp variables for inner loop)
    int ay0 = target_go->y;
    int ay1 = target_go->y + target_go->loc.h;
    int ax0 = target_go->x;
    int ax1 = target_go->x + target_go->loc.w;

    struct particle *p = bullet_go->bullets->next;
    struct particle *previous = bullet_go->bullets;
    struct particle *temp = NULL;

    // no bullets in use
    if( p == NULL ) return;

    do {

        collision = FALSE;

        // Calculate some temp variables of particle sprite.
         bx0 = p->loc.x;
         bx1 = p->loc.x + p->loc.w;
         by0 = p->loc.y;
         by1 = p->loc.y + p->loc.h;

         // Test bounding boxes of player against particle.
         if( bx0 < ax1 ) {
             if( by0 < ay1 ) {
                 if( bx1 > ax0 ) {
                     if( by1 > ay0 ) {

                        if( p->game_effect == DO_DAMAGE ) {
                            damage_go( target_go, p->game_effect_var );
                        }

                        // delete bullet
                        p = delete_bullet( bullet_go, previous );

                        if( target_go == &player ) {
                            score -= 25;
                            if( score < 0 ) score = 0;
                            play_sound( PLAYER_HIT_SOUND );
                        }

                        else if( target_go == &shake ) {
                            score += 50;
                            play_sound( SHAKE_HIT_SOUND );
                        }

                        collision = TRUE;
                    }
                 }
             }
         }  // last if in bounding box collision check logic

        // nothing was deleted
        if( collision == FALSE ) {

            // incriment down the list 1 slot
            previous = p;
            p = p->next;
        }

    } while( p != NULL );  // loop through all bullets in linked list
}



void bullet_goo_collision()
{
     struct particle *a0, *astart, *a;
     struct particle *b0, *bstart, *b;
     int collision = FALSE;
     int ax0, ay0, bx0, by0;
     int ax1, ay1, bx1, by1;
     int count = 0;

     a = player.bullets;
     b = shake.bullets;


     // Elements passed as argumets are default structs that cannot be messed with
     // Skip to first active element.  a0 and b0 still point to previous tho (default structs)
     a0 = a;
     b0 = b;
     astart = a0;
     bstart = b0;

     // If there are no elements (besides the first) in either list there is nothing to collide with.
     if( a0->next == NULL || b0->next == NULL ) return;

     // a,b point to current objects.  a0, b0 point to previous objects.
     a = a0->next;
     b = b0->next;


     // Check the two lists of particles against each other for collision.
     // Delete both both objects in a and b if collision occurs.
     while( a != NULL ) {

        if( b == NULL ) {
            player.active_bullets -= count;
            shake.active_bullets -= count;
            return;
        }

        ax0 = a->x;
        ay0 = a->y;
        ax1 = ax0 + a->loc.w;
        ay1 = ay0 + a->loc.h;

        collision = FALSE;

        while( b != NULL ) {

                  // Load up corners of each sprite's box.
                  bx0 = b->x;
                  by0 = b->y;
                  bx1 = bx0 + b->loc.w;
                  by1 = by0 + b->loc.h;

                  collision = FALSE;

                  // Check location of sprites to each other - success goes to inner loop.
                  if( bx0 < ax1 ) {
                      if( by0 < ay1 ) {
                          if( bx1 > ax0 ) {
                              if( by1 > ay0 ) {
                                  // There is a collision.  Delete the colliding objects.

                                  // reset previous items 'next' pointers to skip over the item we are deleting
                                  a0->next = a->next;
                                  b0->next = b->next;

                                  // delete the two items
                                  if( a == player.bullets ) printf( "FREE'D THE PROTOTYPE PLAYER BULLET IN bullet_goo_collision()-a\n" );
                                  free( a );
                                  if( b == player.bullets ) printf( "FREE'D THE PROTOTYPE PLAYER BULLET IN bullet_goo_collision()-b\n" );
                                  free( b );

                                  score += 10;
                                  play_sound( GOO_HIT_SOUND );

                                  // check for last-in-list condition
                                  // reset last_bullet variable if so
                                  if( a0->next == NULL ) player.last_bullet = a0;
                                  if( b0->next == NULL ) shake.last_bullet = b0;

                                  // Set b = NULL to trigger an exit from inner while loop
                                  b = NULL;

                                  // Load up a with next object in line for checking.
                                  a = a0->next;

                                  // set collision flag
                                  collision = TRUE;
                                  count++;
                              }
                          }
                      }
                  } // end of bounding box check

                  // No collision - next 'b' object to check...
                  if( collision != TRUE ) {

                      // incriment b0 to b
                      b0 = b;
                      b = b->next;
                  }
          }  // 'b' list while loop

          // We reached b = NULL; meaning end of 'b' list or a collision happened.
          // Reset b and b0 to start of list condition.
          b0 = bstart;
          b = b0->next;

          // If a collision occured a and a0 are set correctly already.
          // If it wasn't a collision, b hit the end of the list and no collisions to a occured.
          // incriment a to next item.
          if( collision != TRUE ) {
              a0 = a;
              a = a->next;
          }
     }  // 'a' list while loop

     player.active_bullets -= count;
     shake.active_bullets -= count;
}



void draw_lifemeters()
{
    sdldraw_fillrect( screen, 1, 1,
        ( XREZ / 2 - 2 ) * player.hp / player.max_hp, 12, 0xFFDD3333 );
    sdldraw_fillrect( screen, 1, 14,
        ( XREZ / 2 - 2 ) * player.shield / player.max_shield, 26, 0xFF33DD00 );

    sdldraw_fillrect( screen, XREZ - ( ( XREZ / 2 - 2 ) * shake.hp / shake.max_hp ),
        1, XREZ - 1, 12, 0xFFDD3333 );
    sdldraw_fillrect( screen, XREZ - ( ( XREZ / 2 - 2 ) * shake.shield / shake.max_shield ),
        14, XREZ - 1, 26, 0xFF33DD00 );
}



void draw_score()
{
    sprintf( buffer, "%d", score );
    PutString( screen, 5, 28, buffer );
}


void game_over()
{
    // GUI and event loop
    struct gui_manager guicore;
    int mainmenu_id;

    // GUI button administration details
    int buttonid = FALSE;
    int ms_x, ms_y;
    int highlight = 0;

    // Font SDL surface (SFont.c)
    SDL_Surface *wFont;

    SDL_Event event;


    save_score( score );

    // Load a font and get it ready for use (SFONT requried)
	Font = (SDL_Surface *) SDL_LoadBMP( GAME_FONT_FILE );
	if( ! Font ) {
	    printf( "failed to load font BMP\n" );
	    return -4;
	}
	InitFont( Font );

    // setup GUI gfx elements
	if( init_gui( &guicore ) < 0 ) {
	    fprintf( stderr, "Failed to initialize GUI.\n" );
	    return -1;
	}

	// init the core variables of the gui_manager structure instance
    guicore.winx0 = 0; guicore.winx1 = XREZ; guicore.winy0 = 0; guicore.winy1 = YREZ;
    guicore.first_button = NULL; guicore.first_textbox = NULL; guicore.first_slider = NULL;
    guicore.active_textbox = 0;
    guicore.gui_surface = screen;

    mainmenu_id = add_button( &guicore, 10, 10, 210, 100, "MAIN MENU" );

    SDL_Surface *gameover_gfx = (SDL_Surface *) SDL_LoadBMP( GAMEOVER_GFX_FILE );
	if( ! gameover_gfx ) {
	    printf( "failed to load Game Over BMP\n" );
	    exit( -1 );
	}

    SDL_BlitSurface( gameover_gfx, NULL, screen, NULL );
    shake.loc.x = 532;
    shake.loc.y = 68;
    SDL_BlitSurface( shake.gfx, &shake.anim_rect, screen, &shake.loc );
    draw_gui( &guicore );
    SDL_Flip( screen );


    int done = FALSE;
    while ( ! done ) {

        // wait for an SDL event to process - 0% processor usage sleep/wait
        SDL_WaitEvent( &event );

            switch( event.type ) {

            // exit if the window is closed
            case SDL_QUIT:
            fprintf( stderr, "SDL EVENT: window close button\n" );
                done = TRUE;
                break;

                // check for keypresses
            case SDL_KEYDOWN:

                // exit if ESCAPE is pressed
                if ( event.key.keysym.sym == SDLK_ESCAPE ) {
                    fprintf( stderr, "[ESC] KEY ###################################################\n" );
                    done = TRUE;
                }

                break;

            case SDL_MOUSEBUTTONDOWN:
                if( event.button.button == SDL_BUTTON_LEFT ) {

                    ms_x = event.button.x;
                    ms_y = event.button.y;

                    buttonid = check_buttons( &guicore, ms_x, ms_y );
                    if( buttonid > 0 ) {
                        highlight_button( &guicore, buttonid, HIGHLIGHT );
                        highlight = buttonid;
                    }
                }
                break;

            case SDL_MOUSEBUTTONUP:
                if( event.button.button == SDL_BUTTON_LEFT  ) {

                    ms_x = event.button.x;
                    ms_y = event.button.y;

                    buttonid = check_buttons( &guicore, ms_x, ms_y );
                    if( highlight > 0 ) {
                        highlight_button( &guicore, highlight, CLEAR );
                        highlight = 0;
                    }

                    // quit button
                    if( buttonid == mainmenu_id ) {
                        printf( "QUIT BUTTON\n" );
                        done = TRUE;
                    }

                }  // left button if statement

                break;

            // SDL event type, no match found
            default:
                break;

        } // end switch on SDL event type

        SDL_Flip( screen );

    } //while ! done

    free_gui( &guicore );
    SDL_FreeSurface( gameover_gfx );
}



void you_win()
{
    // GUI and event loop
    struct gui_manager guicore;
    int mainmenu_id;

    // GUI button administration details
    int buttonid = FALSE;
    int ms_x, ms_y;
    int highlight = 0;

    // Font SDL surface (SFont.c)
    SDL_Surface *wFont;

    SDL_Event event;


    save_score( score );

    // Load a font and get it ready for use (SFONT requried)
	Font = (SDL_Surface *) SDL_LoadBMP( GAME_FONT_FILE );
	if( ! Font ) {
	    printf( "failed to load font BMP\n" );
	    return -4;
	}
	InitFont( Font );

    // setup GUI gfx elements
	if( init_gui( &guicore ) < 0 ) {
	    fprintf( stderr, "Failed to initialize GUI.\n" );
	    return -1;
	}

	// init the core variables of the gui_manager structure instance
    guicore.winx0 = 0; guicore.winx1 = XREZ; guicore.winy0 = 0; guicore.winy1 = YREZ;
    guicore.first_button = NULL; guicore.first_textbox = NULL; guicore.first_slider = NULL;
    guicore.active_textbox = 0;
    guicore.gui_surface = screen;

    mainmenu_id = add_button( &guicore, XREZ - 230, YREZ - 110, 210, 100, "MAIN MENU" );

    SDL_Surface *youwin_gfx = (SDL_Surface *) SDL_LoadBMP( WIN_GFX_FILE );
	if( ! youwin_gfx ) {
	    printf( "failed to load Win Screen BMP\n" );
	    exit( -1 );
	}

    SDL_BlitSurface( youwin_gfx, NULL, screen, NULL );
    draw_gui( &guicore );
    SDL_Flip( screen );

    int done = FALSE;
    while ( ! done ) {

        // wait for an SDL event to process - 0% processor usage sleep/wait
        SDL_WaitEvent( &event );

            switch( event.type ) {

            // exit if the window is closed
            case SDL_QUIT:
            fprintf( stderr, "SDL EVENT: window close button\n" );
                done = TRUE;
                break;

                // check for keypresses
            case SDL_KEYDOWN:

                // exit if ESCAPE is pressed
                if ( event.key.keysym.sym == SDLK_ESCAPE ) {
                    fprintf( stderr, "[ESC] KEY ###################################################\n" );
                    done = TRUE;
                }

                break;

            case SDL_MOUSEBUTTONDOWN:
                if( event.button.button == SDL_BUTTON_LEFT ) {

                    ms_x = event.button.x;
                    ms_y = event.button.y;

                    buttonid = check_buttons( &guicore, ms_x, ms_y );
                    if( buttonid > 0 ) {
                        highlight_button( &guicore, buttonid, HIGHLIGHT );
                        highlight = buttonid;
                    }
                }
                break;

            case SDL_MOUSEBUTTONUP:
                if( event.button.button == SDL_BUTTON_LEFT  ) {

                    ms_x = event.button.x;
                    ms_y = event.button.y;

                    buttonid = check_buttons( &guicore, ms_x, ms_y );
                    if( highlight > 0 ) {
                        highlight_button( &guicore, highlight, CLEAR );
                        highlight = 0;
                    }

                    // quit button
                    if( buttonid == mainmenu_id ) {
                        printf( "QUIT BUTTON\n" );
                        done = TRUE;
                    }

                }  // left button if statement

                break;

            // SDL event type, no match found
            default:
                break;

        } // end switch on SDL event type

        SDL_Flip( screen );

    } //while ! done

    SDL_FreeSurface( youwin_gfx );
    free_gui( &guicore );
}



int title_screen()
{
    SDL_Surface *title_gfx = (SDL_Surface *) SDL_LoadBMP( TITLE_GFX_FILE );
	if( ! title_gfx ) {
	    printf( "failed to load Title Screen BMP\n" );
	    return -2;
	}

    // Load a font and get it ready for use (SFONT requried)
	Font = (SDL_Surface *) SDL_LoadBMP( GAME_FONT_FILE );
	if( ! Font ) {
	    printf( "failed to load font BMP\n" );
	    return -4;
	}
	InitFont( Font );

    // GUI and event loop
    struct gui_manager guicore;
    int quit_id, play_id, help_id, scores_id;

    // GUI button administration details
    int buttonid = FALSE;
    int ms_x, ms_y;
    int highlight = 0;

    // Font SDL surface (SFont.c)
    SDL_Surface *Font;

    SDL_Event event;

    // setup GUI gfx elements
	if( init_gui( &guicore ) < 0 ) {
	    fprintf( stderr, "Failed to initialize GUI.\n" );
	    return -1;
	}

	// init the core variables of the gui_manager structure instance
    guicore.winx0 = 0; guicore.winx1 = XREZ; guicore.winy0 = 0; guicore.winy1 = YREZ;
    guicore.first_button = NULL; guicore.first_textbox = NULL; guicore.first_slider = NULL;
    guicore.active_textbox = 0;
    guicore.gui_surface = screen;

    quit_id = add_button( &guicore, 15, 145, 110, 45, "QUIT" );
    play_id = add_button( &guicore, 165, 145, 110, 45, "PLAY" );
    help_id = add_button( &guicore, 315, 145, 110, 45, "HELP" );
    scores_id = add_button( &guicore, 465, 145, 155, 45, "SCORES" );

    SDL_BlitSurface( title_gfx, NULL, screen, NULL );
    draw_gui( &guicore );
    SDL_Flip( screen );


    int done = FALSE;
    while ( ! done ) {

        // wait for an SDL event to process - 0% processor usage sleep/wait
        SDL_WaitEvent( &event );

            switch( event.type ) {

            // exit if the window is closed
            case SDL_QUIT:
            fprintf( stderr, "SDL EVENT: window close button\n" );
                done = TRUE;
                break;

                // check for keypresses
            case SDL_KEYDOWN:

                // exit if ESCAPE is pressed
                if ( event.key.keysym.sym == SDLK_ESCAPE ) {
                    fprintf( stderr, "[ESC] KEY ###################################################\n" );
                    done = TRUE;
                }

                break;

            case SDL_MOUSEBUTTONDOWN:
                if( event.button.button == SDL_BUTTON_LEFT ) {

                    ms_x = event.button.x;
                    ms_y = event.button.y;

                    buttonid = check_buttons( &guicore, ms_x, ms_y );
                    if( buttonid > 0 ) {
                        highlight_button( &guicore, buttonid, HIGHLIGHT );
                        highlight = buttonid;
                    }
                }
                break;

            case SDL_MOUSEBUTTONUP:
                if( event.button.button == SDL_BUTTON_LEFT  ) {

                    ms_x = event.button.x;
                    ms_y = event.button.y;

                    buttonid = check_buttons( &guicore, ms_x, ms_y );
                    if( highlight > 0 ) {
                        highlight_button( &guicore, highlight, CLEAR );
                        highlight = 0;
                    }

                    // quit button
                    if( buttonid == quit_id ) {
                        printf( "QUIT BUTTON\n" );
                        done = TRUE;
                    }

                    if( buttonid == play_id ) {
                        return play_state();
                    }

                    if( buttonid == help_id ) {
                        help_screen();
                        SDL_BlitSurface( title_gfx, NULL, screen, NULL );
                        draw_gui( &guicore );
                        SDL_Flip( screen );
                    }

                    if( buttonid == scores_id ) {
                        score_screen();
                        SDL_BlitSurface( title_gfx, NULL, screen, NULL );
                        draw_gui( &guicore );
                        SDL_Flip( screen );
                    }

                }  // left button if statement

                break;

            // SDL event type, no match found
            default:
                break;

        } // end switch on SDL event type

        SDL_Flip( screen );

    } //while ! done

    SDL_FreeSurface( title_gfx );
    free_gui( &guicore );

    printf( "title screen exit\n" );
    exit( 0 );
}



void score_screen()
{
    // Load a font and get it ready for use (SFONT requried)
	Font = (SDL_Surface *) SDL_LoadBMP( GAME_FONT_FILE );
	if( ! Font ) {
	    printf( "failed to load font BMP\n" );
	    return -4;
	}
	InitFont( Font );

    SDL_Surface *score_gfx = (SDL_Surface *) SDL_LoadBMP( SCORE_GFX_FILE );
	if( ! score_gfx ) {
	    printf( "failed to load Score Screen BMP\n" );
	    return;
	}

    SDL_BlitSurface( score_gfx, NULL, screen, NULL );

    // draw score text from text file data
    FILE *f = fopen( "scores.txt", "r" );
    if( !f ) {
        empty_scores();
        f = fopen( "scores.txt", "r" );
        if( !f ) return;
    }

    int n;
    int score;
    char name[ 80 ];

    for( n = 0; n < 5; n++ ) {
        fscanf( f, "%s\t%ld\n", name, &score );
        sprintf( buffer, "%s    %d", name, score );
        PutString( screen, 65, 150 + 30 * n, buffer );
    }

    for( n = 0; n < 5; n++ ) {
        fscanf( f, "%s\t%ld\n", name, &score );
        sprintf( buffer, "%s    %d", name, score );
        PutString( screen, 350, 150 + 30 * n, buffer );
    }

    fclose( f );

    printf( "score screen waiting for key...\n" );

    SDL_Flip( screen );

    sdl_wait_userinput();

    SDL_FreeSurface( score_gfx );
}



// create default score file
void empty_scores()
{
    FILE *f = fopen( "scores.txt", "w" );

    int n;
    for( n = 0; n < 10; n++ ) {
        fprintf( f, "XXX\t%ld\n", ( 10 - n ) * 100 );
    }

    fclose( f );
}



// doesn't save 3 letter initials yet
// need a place to enter initials too
int save_score( long int m_score )
{
    long int scores[ 10 ];
    int n;
    int slot = -1;

    // draw score text from text file data
    FILE *f = fopen( "scores.txt", "r" );
    if( !f ) {
        empty_scores();
        f = fopen( "scores.txt", "r" );
        if( !f ) return;
    }

    // load scores into array
    for( n = 0; n < 10; n++ ) {
        fscanf( f, "%s\t%ld\n", buffer, &scores[ n ] );
    }

    fclose( f );

    // check for first match >=
    for( n = 0; n < 10; n++ ) {
        if( m_score >= scores[ n ] ) {
            scores[ n ] = m_score;
            slot = n;
            break; // early exit for loop
        }
    }

    if( slot < 0 ) return;

    scores[ slot ] = m_score;

    f = fopen( "scores.txt", "w" );

    for( n = 0; n < 10; n++ ) {
        fprintf( f, "XXX\t%ld\n", scores[ n ] );
    }

    fclose( f );
}



void help_screen()
{
    SDL_Surface *help_gfx = (SDL_Surface *) SDL_LoadBMP( HELP_GFX_FILE );
	if( ! help_gfx ) {
	    printf( "failed to load Help Screen BMP\n" );
	    return;
	}

    SDL_BlitSurface( help_gfx, NULL, screen, NULL );
    SDL_Flip( screen );

    printf( "help screen waiting for key...\n" );

    sdl_wait_userinput();

    SDL_FreeSurface( help_gfx );
}



// SDL Audio mixing callback example code
void mix_audio( void *unused, Uint8 *stream, int len )
{
    int i;
    Uint32 amount;

    for( i = 0; i < NUM_SOUNDS; ++i ) {
        amount = ( sounds[ i ].dlen - sounds[ i ].dpos );
        if ( amount > len ) {
            amount = len;
        }

        SDL_MixAudio( stream, &sounds[ i ].data[ sounds[ i ].dpos ], amount, SDL_MIX_MAXVOLUME );

        sounds[ i ].dpos += amount;
    }
}



// SDL Audio play sound example code
void play_sound( char *file )
{
    int index;
    SDL_AudioSpec wave;
    Uint8 *data;
    Uint32 dlen;
    SDL_AudioCVT cvt;

    /* Look for an empty (or finished) sound slot */
    for ( index=0; index<NUM_SOUNDS; ++index ) {
        if ( sounds[index].dpos == sounds[index].dlen ) {
            break;
        }
    }
    if ( index == NUM_SOUNDS ) {
        return;
    }

    /* Load the sound file and convert it to 16-bit stereo at 22kHz */
    if ( SDL_LoadWAV(file, &wave, &data, &dlen) == NULL ) {
        fprintf(stderr, "Couldn't load %s: %s\n", file, SDL_GetError());
        return;
    }
    SDL_BuildAudioCVT(&cvt, wave.format, wave.channels, wave.freq,
                            AUDIO_S16,   2,             22050);
    cvt.buf = malloc(dlen*cvt.len_mult);
    memcpy(cvt.buf, data, dlen);
    cvt.len = dlen;
    SDL_ConvertAudio(&cvt);
    SDL_FreeWAV(data);

    /* Put the sound data in the slot (it starts playing immediately) */
    if ( sounds[index].data ) {
        free(sounds[index].data);
    }
    SDL_LockAudio();
    sounds[index].data = cvt.buf;
    sounds[index].dlen = cvt.len_cvt;
    sounds[index].dpos = 0;
    SDL_UnlockAudio();
}



void draw_eyelasers()
{
    int n;

    if( laser.status == TRUE ) {
        for( n = 0; n < 50; n++ ) {
            sdl_line( screen, ( shake.x + 15.0 + 4.0 * ( rand() % 10 ) / 10.0 ), shake.y + 24,
                ( laser.x + 4.0 + 16.0 * ( rand() % 10 ) / 10.0 ), laser.y + 5.0,
                0xAADD00DD + (( rand() % 256 ) << 8 ) );
            sdl_line( screen, ( shake.x + 48.0 + 4.0 * ( rand() % 10 ) / 10.0 ), shake.y + 24,
                ( laser.x - 4.0 + 32.0 + 16.0 * ( rand() % 10 ) / 10.0 ), laser.y + 5.0,
                0xAADD00DD + ( ( rand() % 256 ) << 8 ) );
        }
    }
}
