// See license.txt for copyright information and redistribution information

#include "globals.h"
#include "particles.h"

extern int global_particle_indentifier;



int get_next_particle_id()
{
    global_particle_identifier++;
    return global_particle_identifier;
}

void clear_particle_id()
{
    global_particle_identifier = 0;
}



void set_position( struct particle *mp, float x, float y )
{
    mp->x = x;
    mp->y = y;
    mp->loc.x = mp->x;
    mp->loc.y = mp->y;
}

void set_velocity( struct particle *mp, float vx, float vy )
{
    mp->vx = vx;
    mp->vy = vy;
}


int particle_loadgfx( struct particle *mp, const char *filename )
{
    mp->gfx = SDL_LoadBMP( filename );
    if( mp->gfx == NULL ) {
        printf( "Failed to load sprite or particle graphic: %s\n", filename );
        return -1;
    }

    mp->gfx = SDL_DisplayFormat( mp->gfx );
    SDL_SetColorKey( mp->gfx, SDL_SRCCOLORKEY|SDL_RLEACCEL, 0 );  // black is transparent color

    printf( "%s surface loaded.\n", filename );

    // load BMP size into w & h of blitting rectangle
    mp->loc.w = mp->gfx->w;
    mp->loc.h = mp->gfx->h;

    return 0;
}



// create a new particle and load variables and default values
struct particle *new_particle( SDL_Surface *gfxbits, int blitmethod, ai_modes default_ai,
    unsigned long int col_effect, int col_strength, int lifespan )
{
    struct particle *new_p;

    new_p = (struct particle *) malloc( sizeof( struct particle ) );
    if( ! new_p ) return NULL;

    new_p->gfx = gfxbits;
    new_p->current_ai_mode = default_ai;
    new_p->blit_method = blitmethod;
    new_p->game_effect = col_effect;
    new_p->game_effect_var = col_strength;
    new_p->lifespan = lifespan;

    // default values
    new_p->loc.x = 0; new_p->loc.y = 0;

    // get these from the gfx_surface size
    if( gfxbits != NULL ) {
        new_p->loc.w = gfxbits->w;
        new_p->loc.h = gfxbits->h;
    }

    // global unique identifier
    new_p->id = get_next_particle_id();

    // next in linked list
    new_p->next = NULL;

    return new_p;
}



// create a new particle from the values in a current particle
struct particle *copy_particle( struct particle *old_p )
{
    struct particle *new_p;

    new_p = (struct particle *) malloc( sizeof( struct particle ) );
    if( ! new_p ) return NULL;

    new_p->gfx = old_p->gfx;
    new_p->current_ai_mode = old_p->current_ai_mode;
    new_p->blit_method = old_p->blit_method;
    new_p->game_effect = old_p->game_effect;
    new_p->game_effect_var = old_p->game_effect_var;
    new_p->lifespan = old_p->lifespan;

    // default values
    new_p->loc.x = old_p->loc.x;
    new_p->loc.y = old_p->loc.y;

    // get these from the gfx_surface size
    new_p->loc.w = old_p->loc.w;
    new_p->loc.h = old_p->loc.h;

    // global unique identifier
    new_p->id = get_next_particle_id();

    // next in linked list
    new_p->next = NULL;

    return new_p;
}



// Check two particle lists agaist each other for collision.
// TO ADD: if a == b check elements in the list against each other.
int particle_particle_collision( struct particle *a, struct particle *b )
{
     struct particle *a0, *astart;
     struct particle *b0, *bstart;
     int collision = FALSE;
     int ax0, ay0, bx0, by0;
     int ax1, ay1, bx1, by1;
     int count = 0;


     // Elements passed as argumets are default structs that cannot be messed with
     // Skip to first active element.  a0 and b0 still point to previous tho (default structs)
     a0 = a;
     b0 = b;
     astart = a0;
     bstart = b0;

     // If there are no elements (besides the first) in either list there is nothing to collide with.
     if( a0->next == NULL || b0->next == NULL ) return 0;

     // a,b point to current objects.  a0, b0 point to previous objects.
     a = a0->next;
     b = b0->next;


     // Check the two lists of particles against each other for collision.
     // Delete both both objects in a and b if collision occurs.
     while( a != NULL ) {

          if( b == NULL ) return count;

          ax0 = a->x;
          ay0 = a->y;
          ax1 = ax0 + a->loc.w;
          ay1 = ay0 + a->loc.h;

          collision = FALSE;

          while( b != NULL ) {

                  // Load up corners of each sprite's box.
                  bx0 = b->x;
                  by0 = b->y;
                  bx1 = bx0 + b->loc.w;
                  by1 = by0 + b->loc.h;

                  collision = FALSE;

                  // Check location of sprites to each other - success goes to inner loop.
                  if( bx0 < ax1 ) {
                      if( by0 < ay1 ) {
                          if( bx1 > ax0 ) {
                              if( by1 > ay0 ) {
                                  // There is a collision.  Delete the colliding objects.

                                  // reset previous items 'next' pointers to skip over the item we are deleting
                                  a0->next = a->next;
                                  b0->next = b->next;

                                  // delete the two items
                                  free( a );
                                  free( b );

                                  // Set b = NULL to trigger an exit from inner while loop
                                  b = NULL;

                                  // Load up a with next object in line for checking.
                                  a = a0->next;

                                  // set collision flag
                                  collision = TRUE;
                                  count++;
                              }
                          }
                      }
                  } // end of bounding box check

                  // No collision - next 'b' object to check...
                  if( collision != TRUE ) {

                      // incriment b0 to b
                      b0 = b;
                      b = b->next;
                  }
          }  // 'b' list while loop

          // We reached b = NULL; meaning end of 'b' list or a collision happened.
          // Reset b and b0 to start of list condition.
          b0 = bstart;
          b = b0->next;

          // If a collision occured a and a0 are set correctly already.
          // If it wasn't a collision, b hit the end of the list and no collisions to a occured.
          // incriment a to next item.
          if( collision != TRUE ) {
              a0 = a;
              a = a->next;
          }
     }  // 'a' list while loop

     return count;
}



void spark_burst( struct particle *p, int x, int y, float i, float j )
{
     struct particle *stuffit = NULL, *oldptr = NULL;
     int oldx = x;
     int oldy = y;
     int n;

     p->lifespan = SPARKBASELIFE;

     // find end of list to append spark objects to
     stuffit = p;
     while( stuffit != NULL ) {
            oldptr = stuffit;
            stuffit = stuffit->next;
     }

     for( n = 0; n < 50; n++ ) {
          stuffit = (struct particle *) malloc( sizeof( struct particle ) );
          if( ! stuffit ) exit( -1 );

         stuffit->blit_method = p->blit_method;
         stuffit->game_effect = p->game_effect;
         stuffit->game_effect_var = p->game_effect_var;
         stuffit->gfx = p->gfx;
         stuffit->loc.x = 0; stuffit->loc.y = 0; stuffit->loc.w = 10; stuffit->loc.h = 10;
         stuffit->lifespan = p->lifespan + ( rand() % SPARKLIFERND );

         // Random velocities: -1.0 to 1.0, sup-pixel accuracy
         stuffit->vx = 0.25 - (float)( ( rand() % 256 ) / 512.0 );
         stuffit->vy = 0.25 - (float)( ( rand() % 256 ) / 512.0 );
         oldx += i;
         oldy += j;
         stuffit->x = oldx;
         stuffit->y = oldy;

         stuffit->next = NULL;
         oldptr->next = stuffit;
         oldptr = stuffit;
    }
}


// Spark lifetime is set by p.lifespan, as is most of the other basic parameters.
// So setup struct particle p with defaults before calling routine.
// vector (i,j) sets line direction.  It also sets line length (length of vector).
// point (x,y) sets start of spark line (x and y are level coordinates)
// Sparks are created along the line in a hardcoded random zone.
// Sparks lifespan is modified by a random value.
// Sparks are assigned a small random velocity vector.
void spark_line( struct particle *p, int x, int y, float i, float j )
{
     struct particle *stuffit = NULL, *oldptr = NULL;
     int oldx = x;
     int oldy = y;
     int n;

     p->lifespan = SPARKBASELIFE;

     // there may already be active sparks.  find end of list.
     stuffit = p;
     while( stuffit != NULL ) {
         oldptr = stuffit;
         stuffit = stuffit->next;
     }

    // ok now add 50 more sparks.  oldptr is set to the last spark in list.
    // Instead of referencing spark0, it would be good to use oldptr -
    //     different types of sparks could be made other than the global spark0
    for( n = 0; n < 50; n++ ) {
         stuffit = (struct particle *) malloc( sizeof( struct particle ) );
         if( ! stuffit ) exit( -2 );

         stuffit->blit_method = p->blit_method;
         stuffit->game_effect = p->game_effect;
         stuffit->game_effect_var = p->game_effect_var;
         stuffit->gfx = p->gfx;
         stuffit->loc.x = 0; stuffit->loc.y = 0; stuffit->loc.w = 10; stuffit->loc.h = 10;
         stuffit->lifespan = p->lifespan + ( rand() % SPARKLIFERND );

         // Random velocities: -1.0 to 1.0, sup-pixel accuracy
         stuffit->vx = 0.25 - (float)( ( rand() % 256 ) / 512.0 );
         stuffit->vy = 0.25 - (float)( ( rand() % 256 ) / 512.0 );
         oldx += i;
         oldy += j;
         stuffit->x = oldx;
         stuffit->y = oldy;

         stuffit->next = NULL;
         oldptr->next = stuffit;
         oldptr = stuffit;
    }
}



void particle_draw( SDL_Surface *d, struct particle *p )
{
     Uint32 d_offset, s_offset, c0, c1;
     unsigned char r0, g0, b0, r1, g1, b1;
     int m, n;
     Uint32 effects_pitch, gfx_pitch;

     if( p == NULL ) return;

     if( p->blit_method == sprite ) {

         while( p != NULL ) {
                 p->loc.x = p->x - p->loc.w / 2;
                 p->loc.y = p->y - p->loc.h / 2;

                 // blit sprite to effects surface
                 SDL_BlitSurface( p->gfx, NULL, d, &p->loc );

                 p = p->next;
         }

         return;
    }

    else if( p->blit_method == alpha_add ) {

        effects_pitch = d->pitch / 4;
        gfx_pitch = p->gfx->pitch / 4;

        while( p != NULL ) {

            p->loc.x = p->x - p->loc.w / 2; // - 16;
            p->loc.y = p->y - p->loc.h / 2; // - 16;

            if( p->loc.x > 0 && p->loc.x < XREZ - p->loc.w && p->loc.y > 0 && p->loc.y < YREZ - p->loc.h ) {

                // add bit data to effects surface
                for( n = 0; n < p->loc.h; n++ ) {
                     for( m = 0; m < p->loc.w; m++ ) {

                         d_offset = p->loc.x + m + ( n + p->loc.y ) * effects_pitch;

                         s_offset = m + n * gfx_pitch;
                         c1 = ((Uint32 *)p->gfx->pixels)[ s_offset ];
                         r1 = c1 & 0xFF;
                         g1 = ( c1 >> 8 ) & 0xFF;
                         b1 = ( c1 >> 16 ) & 0xFF;

                         c0 = ((Uint32 *)d->pixels)[ d_offset ];
                         r0 = c0 & 0xFF;
                         g0 = ( c0 >> 8 ) & 0xFF;
                         b0 = ( c0 >> 16 ) & 0xFF;

                         r0 += r1 >> 2;
                         b0 += b1 >> 2;
                         g0 += g1 >> 2;

                         ( (Uint32 *) d->pixels )[ d_offset ] = ( r0 | ( g0 << 8 ) | ( b0 << 16 ) );
                     }
                 }
             } // IF p->loc inside screen boundry

             p = p->next;

         } // WHILE p != NULL

         return;

    } // method == ADD
}

