// See license.txt for copyright information and redistribution information

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "SDL.h"
// 32bit ONLY video surfaces
// mix of random useful SDL routines
#include "sdl_draw_core.h"




// From SDL examples page; www.sdllib.org
// Handles 8bit, 16bit, 24bit, and 32bit video surfaces fine.
// Also handles big-endian and little-endian machines.

void putpixel(SDL_Surface *surface, int x, int y, Uint32 pixel)
{
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to set */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

    switch(bpp) {
    case 1:
        *p = pixel;
        break;

    case 2:
        *(Uint16 *)p = pixel;
        break;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
            p[0] = (pixel >> 16) & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = pixel & 0xff;
        } else {
            p[0] = pixel & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = (pixel >> 16) & 0xff;
        }
        break;

    case 4:
        *(Uint32 *)p = pixel;
        break;
    }
}



// from SDL examples pages; www.libsdl.org
// works with all video surface byte sizes and machine endian modes
Uint32 getpixel(SDL_Surface *surface, int x, int y)
{
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to retrieve */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

    switch(bpp) {
    case 1:
        return *p;

    case 2:
        return *(Uint16 *)p;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
            return p[0] << 16 | p[1] << 8 | p[2];
        else
            return p[0] | p[1] << 8 | p[2] << 16;

    case 4:
        return *(Uint32 *)p;

    default:
        return 0;       /* shouldn't happen, but avoids warnings */
    }
}



// an off the cuff drawline routine
// clipping is not handled AT ALL
// independant of video surface structure and machine endian mode (calls putpixel())
void sdl_line( SDL_Surface *s, int x0, int y0, int x1, int y1, Uint32 color )
{
    double dx, dy, d, i, j;
    double x, y;
    int n;

    x = x0;
    y = y0;

    dx = x1 - x0;
    dy = y1 - y0;
    d = sqrt( dx * dx + dy * dy );
    if( d < 1.0 ) return;

    i = dx / d;
    j = dy / d;

    putpixel( s, x0, y0, color );

    for( n = 0; n < d; n++ ) {

        x += i; y += j;

        putpixel( s, (int) x, (int) y, color );
    }
}



// Dump basic SDL surface parameters to stderr.
void inspect_surface( char *name, SDL_Surface *s )
{
     fprintf( stderr, "\nINSPECT SURFACE: %s\n", name );
     fprintf( stderr, "refcount : %d\n", s->refcount );
     fprintf( stderr, "flags : %d\n", s->flags );
     fprintf( stderr, "w : %d\th : %d\n", s->w, s->h );
     fprintf( stderr, "pitch : %d\n", s->pitch );
     //fprintf( stderr, "pixels @ %d\n\n", s->pixels );
}



// Dump pixel format details to stderr.
void inspect_pixelformat( char *name, SDL_PixelFormat *p )
{
     fprintf( stderr, "\nINSPECT PIXEL FORMAT: %s\n", name );
     //fprintf( stderr, "palette mem loc : %d\n", p->palette );
     fprintf( stderr, "bits/pixel / bytes/pixel : %d / %d\n", p->BitsPerPixel, p->BytesPerPixel );
     fprintf( stderr, "Rloss  Gloss  Bloss  Aloss : %d\t\t%d\t\t%d\t\t%d\n", p->Rloss, p->Gloss, p->Bloss, p->Aloss );
     fprintf( stderr, "Rshift Gshift Bshift Ashift: %d\t\t%d\t\t%d\t\t%d\n", p->Rshift, p->Gshift, p->Bshift, p->Ashift );
     fprintf( stderr, "Rmask  Gmask  Bmask  Amask : %x\t%x\t\t%x\t\t%x\n", p->Rmask, p->Gmask, p->Bmask, p->Amask );
     fprintf( stderr, "Colorkey : %d\n", p->colorkey );
     fprintf( stderr, "Alpha : %x\n\n", p->alpha );
}




// Dump information about the "best" video mode available to stderr.
void inspect_video( void )
{
     SDL_VideoInfo *info;                  // details of supported hardware video functions
     SDL_Rect **rect_ptrs;                 // list of pointers to supported screen sizes
     SDL_PixelFormat format32 = { NULL, 32, 4, 0, 0, 0, 8, 16, 8, 0, 0, 0xFF0000, 0xFF00, 0xFF, 0, 0, 0 };
     char buffer[ 80 ];
     int i;


     info = (SDL_VideoInfo *) SDL_GetVideoInfo();

     fprintf( stderr, "-----------------------------------------------------\n" );
     SDL_VideoDriverName( buffer, 80 );
     fprintf( stderr, "Video Driver: %s\n", buffer );
     fprintf( stderr, "SW = software SDL surface (sys mem)\nHW = hardware SDL surface (video mem)\n" );
     if( info->hw_available ) fprintf( stderr, "HW Surf: YES\n" );
     else fprintf( stderr, "HW Surf: NO\n" );
     if( info->wm_available ) fprintf( stderr, "Window Manager: YES\n" );
     else fprintf( stderr, "Window Manager: NO\n" );
     if( info->blit_hw ) fprintf( stderr, "HW->HW Blits Accel'ed: YES\n" );
     else fprintf( stderr, "HW->HW BLits Accel'ed: NO\n" );
     if( info->blit_hw_CC ) fprintf( stderr, "HW->HW Colorkey Blits Accel'ed: YES\n" );
     else fprintf( stderr, "HW->HW Colorkey Blits Accel'ed: NO\n" );
     if( info->blit_hw_A ) fprintf( stderr, "HW->HW Alpha Blits Accel'ed: YES\n" );
     else fprintf( stderr, "HW->HW Alpha Blits Accel'ed: NO\n" );
     if( info->blit_sw ) fprintf( stderr, "SW->HW Blits Accel'ed: YES\n" );
     else fprintf( stderr, "SW->HW BLits Accel'ed: NO\n" );
     if( info->blit_sw_CC ) fprintf( stderr, "SW->HW Colorkey Blits Accel'ed: YES\n" );
     else fprintf( stderr, "SW->HW Colorkey BLits Accel'ed: NO\n" );
     if( info->blit_sw_A ) fprintf( stderr, "SW->HW Alpha Blits Accel'ed: YES\n" );
     else fprintf( stderr, "SW->HW Alpha BLits Accel'ed: NO\n" );
     if( info->blit_fill ) fprintf( stderr, "Color Fills Accel'ed: YES\n" );
     else fprintf( stderr, "Color Fills Accel'ed: NO\n" );
     fprintf( stderr, "HW Video Memory (kb): %d\n", info->video_mem );

     inspect_pixelformat( "Current/Best Pixel Format", info->vfmt );

     fprintf( stderr, "Available FULLSCREEN 32bit RGB HARDWARE screen sizes:\n" );
     rect_ptrs = SDL_ListModes( &format32, SDL_FULLSCREEN|SDL_HWSURFACE );
     if( rect_ptrs == 0 ) fprintf( stderr, "No 32bit RGB hardware modes available.\n" );
     if( rect_ptrs == (SDL_Rect **) -1 ) fprintf( stderr, "No restrictions on screen size for 32bit RGB modes.\n" );
     else {
          fprintf( stderr, "32bit RGB available but restricted.  Usable screen sizes are:\n" );
          for( i = 0; rect_ptrs[ i ] != 0; i++ )
               fprintf( stderr, "%d: %d x %d\n", i, rect_ptrs[ i ]->w, rect_ptrs[ i ]->h );
     }

     fprintf( stderr, "-----------------------------------------------------\n\n" );
}



// Draw a rotated and scaled version of [*src] to [*dest]
// [angle] = rotation angle in degrees, (w,h) = width and height of source
// [s] = scale factor, (xcd,ycd) = center of destination (?)
// 32bit color
void sdldraw_surfrot2( SDL_Surface *src, SDL_Surface *dest, int w, int h, int xcd, int ycd, int angle, float s )
{
     Uint32 color = 0;                              // 32bit color dump bucket
     Uint32 offset;                                 // pixel offset in an SDL_Surface->pixels[]

     float t;                                       // rotation angle in radians
     float xo, yo;                                  // source pixmap location w/ sub-pixel accuracy (float)
     float xstart, ystart;                          // source pixmap location of start of each scan line
     float i_x, i_y, j_x, j_y;                      // source pixmap unit vectors after 2D rotation & scale

     int x, y;                                      // destination pixmap pixel location (inner loops)
     int xi, yi;                                    // (xo,yo) point cast into integers
     int xso, yso;                                  // Upper left corner in destination pixmap
     int pdest, psrc;                               // pixels[] offset cast into integers after float calc.'s



     t = angle * PI / 180.0;                        // rotation angle converted to radians

     // Calculate source pixmap unit vector components based on 2d rotation transform.
     i_x = cos( t ) / s;
     i_y = -sin( t ) / s;
     j_x = sin( t ) / s;
     j_y = cos( t ) / s;

     // Calculate starting coordinates based on center of source pixmap & i,j transform vectors
     xstart = w / 2 - s * i_x * w / 2 - s * j_x * h / 2;
     ystart = h / 2 - s * i_y * w / 2 - s * j_y * h / 2;

     // Calculate upper left corner point in destination pixmap
     xso = xcd - s * w / 2;
     yso = ycd - s * h / 2;

     // Calculate dead memory area after each scanline in source and destination SDL_Surfaces
     // SDL_SUrface.pitch is in units of bytes.  We are writting 4 bytes at a time.  Divide by 4.
     pdest = dest->pitch / 4;
     psrc = src->pitch / 4;

     SDL_LockSurface( dest );

     // Loop through vertical scanlines in destination pixmap.
     for( y = 0; y < h * s; y++ ) {

          // Set the source pixmap coordinate to the calc'ed start of source scanline.
          xo = xstart;
          yo = ystart;

          // Loop through horizontal pixels in destination pixmap.
          for( x = 0; x < w * s; x++ ) {

               // Truncate to integers.  Cleaner way to do this?  Besides fixed point math?
               xi = xo;
               yi = yo;

               // Clip source pixmap coordinate to source pixmap boundries.  Transparent if outside.
               if( xi < 0 || xi > w - 1 || yi < 0 || yi > h - 1 ) color = 0;

               // Get pixel color from source pixmap using direct pixel access and surface locking.
               else {
                    offset = xi + yi * psrc;
                    SDL_LockSurface( src );
                    color = ((Uint32 *)src->pixels)[ offset ];
                    SDL_UnlockSurface( src );
               }

               // Write pixel color to destination pixmap.
               offset = xso + x + ( yso + y ) * pdest;
               ( (Uint32 *) dest->pixels )[ offset ] = color;

               // Incriment source pixmap coordinate along the i unit vector
               xo += i_x;
               yo += i_y;
          }

          // Incriment source pixmap coordiante along the j unit vector
          xstart += j_x;
          ystart += j_y;
     }
     SDL_UnlockSurface( dest );
}



// 32bit color, 480 vertical lines
// Draw vector starting at screen center in the direction of [angle]
//    Offset start of draw from center by (dcx,dcy).  Draw length [r]
// Display hardware coordinate system
void sdldraw_polarvec( SDL_Surface *s, Uint32 color, int dcx, int dcy, int r, float angle )
{
     int i;
     Uint32 offset;
     float ix, iy, fx, fy;

     Uint32 max = 480 * s->pitch/4;

     ix = cos( angle * ANGLE_REZ );
     iy = sin( angle * ANGLE_REZ );
     fx = 0;
     fy = 0;

     SDL_LockSurface( s );
     for( i = 0; i < r; i++ ) {
          fx += ix; fy += iy;
          offset = (int) fx + dcx + ( (int) fy + dcy ) * ( s->pitch >> 2 );
          if( offset < max ) ( (Uint32 *) s->pixels )[ offset ] = color;
     }
     SDL_UnlockSurface( s );
}


// 32bit color, 480 vert lines
// Draw cartesian vector starting at screen center (?) in the direction of (x,y)
//    Offset start of vector draw by (dcx,dcy) from screen center.
//    [scale] determines length - vector is not properly normalized...
// Display hardware coordinate system
void sdldraw_cartvec( SDL_Surface *s, Uint32 color, int dcx, int dcy, float x, float y, float scale )
{
    int i;
    float ix, iy, fx, fy, r, sx, sy;
    Uint32 offset;

    Uint32 max = 480 * s->pitch/4;


    ix = x / sqrt( x * x + y * y );
    iy = y / sqrt( x * x + y * y );
    fx = 0;
    fy = 0;
    r = sqrt( x * x + y * y ) * scale;

     SDL_LockSurface( s );
     for( i = 0; i < r; i++ ) {
          fx += ix; fy += iy;
          sx = fx + dcx;
          sy = fy + dcy;
          offset = (int) sx + (int) sy * ( s->pitch >> 2 );
          if( offset < max ) ( (Uint32 *) s->pixels )[ offset ] = color;
     }
     SDL_UnlockSurface( s );
}



// Specialized version of draw_cart_vector for player velocity display
// 32bit color, 480 vertical lines
// Display hardware coordinate system
void sdldraw_ijvec( SDL_Surface *s, Uint32 color, float px, float py, float vx, float vy, float scale, int skip )
{
    int i;
    float ix, iy, fx, fy, r, sx, sy;
    Uint32 offset;
    Uint32 max = 480 * s->pitch/4;


    ix = vx / sqrt( vx * vx + vy * vy );
    iy = vy / sqrt( vx * vx + vy * vy );
    fx = skip * ix;
    fy = skip * iy;
    r = sqrt( vx * vx + vy * vy ) * scale;

     SDL_LockSurface( s );
     for( i = 0; i < r; i++ ) {
          fx += ix; fy += iy;
          sx = fx + px;
          sy = fy + py;
          offset = (int) sx + (int) sy * ( s->pitch >> 2 );
          if( offset < max ) ( (Uint32 *) s->pixels )[ offset ] = color;
     }
     SDL_UnlockSurface( s );
}



// 32bit color filled rectangle
void sdldraw_fillrect( SDL_Surface *s, int x0, int y0, int x1, int y1, Uint32 color )
{
     int i, j;

     SDL_LockSurface( s );

     for( i = x0; i < x1; i++ ) {
          for( j = y0; j < y1; j++ ) {
               ( (Uint32 *) s->pixels )[ ( i + j * s->pitch/4 ) ] = color;
          }
     }

     SDL_UnlockSurface( s );
}


// blit one surf to another, scaling to fill dest size
// 32bit only
void sdldraw_blitscale( SDL_Surface *s, SDL_Surface *d, int x0, int y0, int x1, int y1 )
{
    int i, j;
    float x_inc, y_inc;
    float xs = 0, ys = 0;
    Uint32 color;


    x_inc = ( (float)( s->w ) / (float) ( x1 - x0 ) );
    y_inc = ( (float)( s->h ) / (float) ( y1 - y0 ) );
    //fprintf( stderr, "%f, %f\n", x_inc, y_inc );

    SDL_LockSurface( s );

    for( j = y0; j < y1; j++ ) {
        for( i = x0; i < x1; i++ ) {

            color = getpixel( s, xs, ys );
            putpixel( d, i, j, color );

            xs += x_inc;
        }

        ys += y_inc;
        xs = 0;
    }

    SDL_UnlockSurface( s );
}
