// See license.txt for copyright information and redistribution information

#include <stdlib.h>
#include <string.h>
#include "SFont.h"
#include "sdl_gui_core.h"
#include "sdl_draw_core.h"



// ------------------------------------------------------------------
// Load the bitmap information for GUI element drawing
// ------------------------------------------------------------------
int init_gui( struct gui_manager *g )
{
	g->button_gfx = (SDL_Surface *) SDL_LoadBMP( BUTTON_GFX );
	if( ! g->button_gfx ) {
	    fprintf( stderr, "failed to load button graphic\n" );
	    return -1;
	}
	SDL_SetColorKey( g->button_gfx, SDL_SRCCOLORKEY|SDL_RLEACCEL, 0 );  // 0 transparent & accelerated
	g->button_gfx = SDL_DisplayFormat( g->button_gfx );


	g->button_active_gfx = (SDL_Surface *) SDL_LoadBMP( BUTTON_ACTIVE_GFX );
	if( ! g->button_active_gfx ) {
	    fprintf( stderr, "failed to load button active graphic\n" );
	    return -1;
	}
	SDL_SetColorKey( g->button_active_gfx, SDL_SRCCOLORKEY|SDL_RLEACCEL, 0 );  // 0 transparent & accelerated
	g->button_active_gfx = SDL_DisplayFormat( g->button_active_gfx );

	g->textbox_gfx = (SDL_Surface *) SDL_LoadBMP( TEXTBOX_GFX );
	if( ! g->textbox_gfx ) {
	    fprintf( stderr, "failed to load text box graphic\n" );
	    return -1;
	}
	SDL_SetColorKey( g->textbox_gfx, SDL_SRCCOLORKEY|SDL_RLEACCEL, 0 );  // 0 transparent & accelerated
	g->textbox_gfx = SDL_DisplayFormat( g->textbox_gfx );

	g->pbarfill_gfx = (SDL_Surface *) SDL_LoadBMP( PBOXFILL_GFX );
	if( ! g->pbarfill_gfx ) {
	    fprintf( stderr, "failed to load progress bar fill graphic\n" );
	    return -1;
	}
	SDL_SetColorKey( g->pbarfill_gfx, SDL_SRCCOLORKEY|SDL_RLEACCEL, 0 );  // 0 transparent & accelerated
	g->pbarfill_gfx = SDL_DisplayFormat( g->pbarfill_gfx );


	g->first_button = NULL;
    g->first_textbox = NULL;
    g->first_slider = NULL;
    g->progbar = NULL;
    g->first_textblock = NULL;
    g->fonth = 20;
    g->fontw = 10;

    return 0;
}

// ------------------------------------------------------------------
// Draw all gui elements
// ------------------------------------------------------------------
void draw_gui( struct gui_manager *gm )
{
    draw_buttons( gm );
    draw_sliders( gm );
    draw_textboxes( gm );
    draw_textblocks( gm );
}

// ------------------------------------------------------------------
// Free everything dynamic in the gui_manager
// ------------------------------------------------------------------
void free_gui( struct gui_manager *gm )
{
    free_sliders( gm );
    free_buttons( gm );
    free_textboxes( gm );
    free_progressbar( gm );
    free_textblocks( gm );
}

// ------------------------------------------------------------------
// draw the progress bar with 0 to 1.0 fraction of area filled
// ------------------------------------------------------------------
void draw_progressbar( struct gui_manager *g, float xfill )
{
    struct progressbar *b;
    char buffer[ 256 ];


    if( !g ) return;
    b = g->progbar;
    if( !b ) return;
    if( xfill < 0 || xfill > 1.0 ) return;

    sdldraw_blitscale( g->button_gfx, g->gui_surface, b->x, b->y,
        b->x + b->w, b->y + b->h );

    sdldraw_blitscale( g->pbarfill_gfx, g->gui_surface, b->x + 4, b->y + 4,
        b->x + ( ( b->w - 8 ) * xfill ), b->y + b->h - 8 );

    sdldraw_fillrect( g->gui_surface, b->x + 4, b->y + b->h + 4, b->x + b->w, b->y + b->h + 25, 0x00000000 );

    sprintf( buffer, "%s", b->text );
    PutString( g->gui_surface, b->x + 4, b->y + b->h + 4, buffer );

    SDL_Flip( g->gui_surface );
}

// ------------------------------------------------------------------
// initialize the progress bar.  Only 1 allowed per GUI
// ------------------------------------------------------------------
void init_progressbar( struct gui_manager *g, int x0, int y0, int w, int h, char *t )
{
    struct progressbar *newpb;


    newpb = (struct progressbar *) malloc( sizeof( struct progressbar ) );
    if( !newpb ) return;

    g->progbar = newpb;
    newpb->x = x0;
    newpb->y = y0;
    newpb->w = w;
    newpb->h = h;

    strcpy( newpb->text, t );

    return;
}

// ------------------------------------------------------------------
// set the text of the progress bar
// ------------------------------------------------------------------
void settext_progressbar( struct gui_manager *g, char *t )
{
    strcpy( g->progbar->text, t );
}

// ------------------------------------------------------------------
// Free the progress bar
// ------------------------------------------------------------------
void free_progressbar( struct gui_manager *g )
{
    if( !g ) return;
    if( !g->progbar ) return;


    if( g->gui_surface ) sdldraw_fillrect( g->gui_surface, g->progbar->x, g->progbar->y,
        g->progbar->x + g->progbar->w, g->progbar->y + g->progbar->h, 0x00000000 );

    free( g->progbar );
}

// ------------------------------------------------------------------
// Draw all button objects in a linked list
// ------------------------------------------------------------------
void draw_buttons( struct gui_manager *g )
{
    struct button *b = NULL;
    char buffer[ 80 ];


    if( !g ) return;

    b = g->first_button;
    if( !b ) return;

    do {
        sdldraw_blitscale( g->button_gfx, g->gui_surface, b->x, b->y, b->x + b->w, b->y + b->h );

        sprintf( buffer, b->text );
        PutString( g->gui_surface, b->x + 6, b->y + 4, buffer );

    } while( ( b = b->next ) );
}

// ------------------------------------------------------------------
// Search linked list of buttons for ID, return pointer to it
// ------------------------------------------------------------------
struct button *get_button( struct gui_manager *gm, int id )
{
    struct button *b;

    // either the id or init pointer is bad or empty, return empty
    if( !gm || id == 0 ) return NULL;
    if( !gm->first_button ) return NULL;

    b = gm->first_button;

    do {
        if( b->id == id ) return b;
    } while( ( b = b->next ) );

    // no match found, return empty
    fprintf( stderr, "No matching button to input ID %d\n", id );
    return NULL;
}

// ------------------------------------------------------------------
// Search linked list of text boxes for ID, return pointer to it
// ------------------------------------------------------------------
struct text_box *get_textbox( struct gui_manager *gm, int id )
{
    struct text_box *b;

    // either the id or init pointer is bad or empty, return empty
    if( !gm || id == 0 ) {
        fprintf(stderr, "bad gui manager pointer or bad id\n" );
        return NULL;
    }

    if( !gm->first_textbox ) {
        fprintf( stderr, "empty text box list in gui manager\n" );
        return NULL;
    }

    b = gm->first_textbox;

    do {
        if( b->id == id ) return b;
    } while( ( b = b->next ) );

    // no match found, return empty
    fprintf( stderr, "No matching text box to input ID %d\n", id );
    return NULL;
}

// ------------------------------------------------------------------
// Search linked list of text blocks for ID, return pointer to it
// ------------------------------------------------------------------
struct text_block *get_textblock( struct gui_manager *gm, int id )
{
    struct text_block *b;

    // either the id or init pointer is bad or empty, return empty
    if( !gm || id == 0 ) {
        fprintf(stderr, "bad gui manager pointer or bad ID\n" );
        return NULL;
    }

    if( !gm->first_textblock ) {
        fprintf( stderr, "empty text block list in gui manager\n" );
        return NULL;
    }

    b = gm->first_textblock;

    do {
        if( b->id == id ) return b;
    } while( ( b = b->next ) );

    // no match found, return empty
    fprintf( stderr, "No matching text block to input ID %d\n", id );
    return NULL;
}

// ------------------------------------------------------------------
// return the pointer to the last button in the gui manager list
// ------------------------------------------------------------------
struct button *get_last_button( struct gui_manager *gm )
{
    struct button *b;

    // double check gui manager pointers
    if( !gm ) return NULL;
    if( !gm->first_button ) return NULL;

    // load the first in the list
    b = gm->first_button;

    while( b->next ) {
        b = b->next;
    }

    return b;
}

// ------------------------------------------------------------------
// return the pointer to the last text block in the gui manager list
// ------------------------------------------------------------------
struct text_block *get_last_textblock( struct gui_manager *gm )
{
    struct text_block *b;

    // double check gui manager pointers
    if( !gm ) return NULL;
    if( !gm->first_textblock ) return NULL;

    // load the first in the list
    b = gm->first_textblock;

    while( b->next ) {
        b = b->next;
    }

    return b;
}

// ------------------------------------------------------------------
// return the pointer to the last text box in the gui manager list
// ------------------------------------------------------------------
struct text_box *get_last_textbox( struct gui_manager *gm )
{
    struct text_box *b;

    // double check gui manager pointers
    if( !gm ) return NULL;
    if( !gm->first_textbox ) return NULL;

    // load the first in the list
    b = gm->first_textbox;

    while( b->next ) {
        b = b->next;
    }

    return b;
}

// ------------------------------------------------------------------
// return the pointer to the last button in the gui manager list
// ------------------------------------------------------------------
struct slider *get_last_slider( struct gui_manager *gm )
{
    struct slider *b;

    // double check gui manager pointers
    if( !gm ) return NULL;
    if( !gm->first_slider ) return NULL;

    // load the first in the list
    b = gm->first_slider;

    while( b->next ) {
        b = b->next;
    }

    return b;
}

// ------------------------------------------------------------------
// Highlight or clear button gfx
// ------------------------------------------------------------------
void highlight_button( struct gui_manager *gm, int id, int state )
{
    char buffer[ 80 ];
    struct button *b = NULL;

    if( !gm || id == 0 ) return;

    // find pointer to the right button
    b = get_button( gm, id );
    if( !b ) return;

    switch( state ) {
        case HIGHLIGHT:
            sdldraw_blitscale( gm->button_active_gfx, gm->gui_surface, b->x, b->y, b->x + b->w, b->y + b->h );
            break;

        case CLEAR:
            sdldraw_blitscale( gm->button_gfx, gm->gui_surface, b->x, b->y, b->x + b->w, b->y + b->h );
            break;

        default:
            break;
    }

    sprintf( buffer, b->text );
    PutString( gm->gui_surface, b->x + 6, b->y + 4, buffer );
}

// ------------------------------------------------------------------
// Draw all text boxes objects in a linked list
// ------------------------------------------------------------------
void draw_textboxes( struct gui_manager *g )
{
    struct text_box *b = NULL;
    char buffer[ 256 ];
    int twidth = 0;


    if( !g ) return;

    b = g->first_textbox;
    if( !b ) return;

    do {
        sdldraw_blitscale( g->textbox_gfx, g->gui_surface, b->x, b->y, b->x + b->w, b->y + b->h );

        sprintf( buffer, b->text );
        twidth = TextWidth( buffer );
        PutString( g->gui_surface, b->x + 2, b->y + 2, buffer );

        sprintf( buffer, b->data );
        PutString( g->gui_surface, b->x + twidth + 2, b->y + 2, buffer );

    } while( ( b = b->next ) );
}

// ------------------------------------------------------------------
// Draw all text blocks objects in a linked list
// ------------------------------------------------------------------
void draw_textblocks( struct gui_manager *g )
{
    struct text_block *b = NULL;
    char buffer[ 256 ];
    int twidth = 0;
    int i;


    if( !g ) return;

    b = g->first_textblock;
    if( !b ) return;

    do {
        sdldraw_blitscale( g->textbox_gfx, g->gui_surface, b->x, b->y, b->x + b->w, b->y + b->h );

        // start drawing text lines starting at line #start and going to end of text block height
        for( i = b->start; i < b->theight + b->start; i++ ) {
            if( i <= b->end ) {
                sprintf( buffer, b->line[ i ] );
                twidth = TextWidth( buffer );
                PutString( g->gui_surface, b->x + 6, b->y + 6 + i * g->fonth, buffer );
            }
        }

    } while( ( b = b->next ) );
}

// ------------------------------------------------------------------
// Check to see if supplied (x,y) coordinates are inside any buttons
// Return button id or FALSE
// ------------------------------------------------------------------
int check_buttons( struct gui_manager *gm, int x, int y )
{
    struct button *b;

    if( !gm ) return FALSE;
    if( !gm->first_button ) return FALSE;

    b = gm->first_button;

    do {
        if( x > b->x && x < b->x + b->w && y > b->y && y < b->y + b->h ) return b->id;
    } while( ( b = b->next ) );
    return FALSE;
}

// ------------------------------------------------------------------
// Check to see if supplied (x,y) coordinates are inside text boxes
// Return button id or FALSE
// ------------------------------------------------------------------
int check_textboxes( struct gui_manager *gm, int x, int y )
{
    struct text_box *b;

    if( !gm ) return FALSE;
    if( !gm->first_textbox ) return FALSE;

    b = gm->first_textbox;

    do {
        if( x > b->x && x < b->x + b->w && y > b->y && y < b->y + b->h ) return b->id;
    } while( ( b = b->next ) );
    return FALSE;
}

// ------------------------------------------------------------------
// Return the id of a slider if it is clicked
// ------------------------------------------------------------------
int check_sliders( struct gui_manager *gm, int x, int y )
{
    struct slider *l;

    if( !gm ) return FALSE;
    if( !gm->first_slider ) return FALSE;

    l = gm->first_slider;

    do {
        if( x > l->x && x < l->x + l->w && y > ( l->y - 5 ) && y < l->y + 5 ) {
            l->data = ( x - l->x ) * ( 100.0 / l->w );
            return l->id;
        }
    } while( ( l = l->next ) );
    return FALSE;
}

// ------------------------------------------------------------------
// Initialize a button object, attach to linked list
// ------------------------------------------------------------------
int add_button( struct gui_manager *gm, int x, int y, int w, int h, char *text )
{
    struct button *newb, *b;

    newb = malloc( sizeof( struct button ) );
    if( !newb ) return 0;

    if( gm->first_button ) {
        b = get_last_button( gm );
        b->next = newb;
        newb->id = b->id + 1;
    }
    else {
        gm->first_button = newb;
        newb->id = 101;
    }

    newb->x = x;
    newb->y = y;
    newb->w = w;
    newb->h = h;
    newb->next = NULL;
    strcpy( newb->text, text );

    return newb->id;
}

// ------------------------------------------------------------------
// Initialize a text block object, attach to linked list
// ------------------------------------------------------------------
int add_textblock( struct gui_manager *gm, int x, int y, int w, int h )
{
    struct text_block *newb, *b;
    int i;


    newb = malloc( sizeof( struct text_block ) );
    if( !newb ) return 0;

    if( gm->first_textblock ) {
        b = get_last_textblock( gm );
        b->next = newb;
        newb->id = b->id + 1;
    }
    else {
        gm->first_textblock = newb;
        newb->id = 101;
    }

    newb->x = x;
    newb->y = y;
    newb->w = w;
    newb->h = h;
    newb->next = NULL;
    newb->twidth = w / gm->fontw - 2;
    newb->theight = h / gm->fonth;
    newb->start = 0;
    newb->end = -1;

    // malloc 256 lines of twidth characters
    for( i = 0; i < 256; i++ ) {
        newb->line[ i ] = malloc( sizeof( char ) * 256 );
        if( !newb->line[ i ] ) {
            fprintf( stderr, "Failed to allocate some memory for a text block GUI element\n" );
            return -1;
        }
        // set each line to a end-of-string character - initialzed empty
        *( newb->line[ i ] ) = '\0';
    }

    return newb->id;
}

// ------------------------------------------------------------------
// Reset text block to initial conditions
// ------------------------------------------------------------------
void clear_textblock( struct gui_manager *g, int id )
{
    int i;
    struct text_block *t;


    if( !g ) return;
    if( id <= 0 ) return;

    t = get_textblock( g, id );

    // set each line starting char to end-of-string char
    for( i = 0; i < 256; i++ ) {
        *( t->line[ i ] ) = '\0';
    }

    // line counters reset
    t->start = 0;
    t->end = -1;
}

// ------------------------------------------------------------------
// Draw a text entry cursor in a text entry box that is active
// ------------------------------------------------------------------
void draw_cursor( SDL_Surface *s, struct text_box *b )
{
    int twidth;

    twidth = TextWidth( b->data ) + TextWidth( b->text );
    sdldraw_fillrect( s, b->x + twidth, b->y + 12, b->x + twidth + 12, b->y + 16, 0xFFFFAA00 );
}

// ------------------------------------------------------------------
// Initialize a text box object, attach to linked list
// ------------------------------------------------------------------
int add_textbox( struct gui_manager *gm, int x, int y, int w, int h, char *text, char *data )
{
    struct text_box *newt, *t;

    newt = malloc( sizeof( struct text_box ) );
    if( !newt ) return 0;

    if( gm->first_textbox ) {
        t = get_last_textbox( gm );
        t->next = newt;
        newt->id = t->id + 1;
    }
    else {
        gm->first_textbox = newt;
        newt->id = 101;
    }

    newt->x = x;
    newt->y = y;
    newt->w = w;
    newt->h = h;
    newt->next = NULL;
    strcpy( newt->text, text );
    strcpy( newt->data, data );

    return newt->id;
}

// ------------------------------------------------------------------
// Free all buttons in a linked list
// ------------------------------------------------------------------
void free_buttons( struct gui_manager *gm )
{
    struct button *next, *l;

    if( !gm ) return;
    if( !gm->first_button ) return;

    l = gm->first_button;

    // Step down the linked list, deallocating elements as we go.
    while( l ) {
        fprintf( stderr, "Freeing button id %d\n", l->id );
        next = l->next;
        free( l );
        l = next;
    }
}

// ------------------------------------------------------------------
// Free all text blocks in a linked list
// ------------------------------------------------------------------
void free_textblocks( struct gui_manager *gm )
{
    struct text_block *next, *l;
    int i;


    if( !gm ) return;
    if( !gm->first_textblock ) return;

    l = gm->first_textblock;

    // Step down the linked list, deallocating elements as we go.
    while( l ) {
        fprintf( stderr, "Freeing text block id %d\n", l->id );

        for( i = 0; i < 256; i++ ) if( l->line[ i ] ) free( l->line[ i ] );

        next = l->next;
        free( l );
        l = next;
    }
}

// ------------------------------------------------------------------
// Free a text entry field linked list
// ------------------------------------------------------------------
void free_textboxes( struct gui_manager *gm )
{
    struct text_box *next, *l;

    if( !gm ) return;
    if( !gm->first_textbox ) return;

    l = gm->first_textbox;

    // Step down the linked list, deallocating elements as we go.
    while( l ) {
        fprintf( stderr, "Freeing text box id %d\n", l->id );
        next = l->next;
        free( l );
        l = next;
    }
}

// ------------------------------------------------------------------
// Draw all slider objects in linked list
// ------------------------------------------------------------------
void draw_sliders( struct gui_manager *g )
{
    struct slider *l = NULL;
    float incr;


    if( !g ) return;
    l = g->first_slider;

    if( !l ) return;

    do {
        sdldraw_fillrect( g->gui_surface, l->x, l->y - 5, l->x + l->w, l->y + 5, 0xFF000000 );
        sdldraw_fillrect( g->gui_surface, l->x, l->y - 2, l->x + l->w, l->y + 2, 0xFFFFFFFF );

        incr = l->w / 100.0;

        sdldraw_fillrect( g->gui_surface, l->x + l->data * incr, l->y - 5,
                          l->x + l->data * incr + 4, l->y + 5, 0xFFFF00FF );

    } while( ( l = l->next ) );
}

// ------------------------------------------------------------------
// Init a new slider object
// ------------------------------------------------------------------
int add_slider( struct gui_manager *gm, int x, int y, int w, int data )
{
    struct slider *newl, *l;

    newl = malloc( sizeof( struct slider ) );
    if( !newl ) return 0;

    if( gm->first_slider ) {
        l = get_last_slider( gm );
        l->next = newl;
        newl->id = l->id + 1;
    }
    else {
        gm->first_slider = newl;
        newl->id = 101;
    }

    newl->x = x;
    newl->y = y;
    newl->w = w;
    newl->next = NULL;
    newl->data = data;

    return newl->id;
}


// ------------------------------------------------------------------
// Free all sliders in a linked list
// ------------------------------------------------------------------
void free_sliders( struct gui_manager *gm )
{
    struct slider *next, *l;

    if( !gm ) return;
    if( !gm->first_slider ) return;

    l = gm->first_slider;

    // Step down the linked list, deallocating elements as we go.
    while( l ) {
        fprintf( stderr, "Freeing slider id %d\n", l->id );
        next = l->next;
        free( l );
        l = next;
    }
}

// ------------------------------------------------------------------
// add a character to the end of a string, move /0 character
// ------------------------------------------------------------------
void addchar( char *string, char c )
{
    int length;


    length = strlen( string );

    if( length > 78 ) return;

    string[ length ] = c;
    string[ length + 1 ] = '\0';
}

// ------------------------------------------------------------------
// delete a character from the end of a string, move /0 character
// ------------------------------------------------------------------
void backspacechar( char *string )
{
    int length;

    length = strlen( string );
    if( length < 1 ) return;

    string[ length - 1 ] = '\0';
}

// ------------------------------------------------------------------
// add a line to the text block buffer
// have to break up the line if it is too long and wrap it
// ------------------------------------------------------------------
void tb_addline( struct text_block *b, char *newline )
{
    if( !b ) return;

    // incr the end line count - special bounding rules
    if( b->end < 0 ) b->end = 0;
    else if( b->end < 255 ) b->end++;
    else b->end = 255;

    // strip any newline characters from input line
    strip_newline( newline );

    // while newline > twidth
    while( newline != NULL ) {
        fprintf( stderr, "remainder: %s\n", newline );
        newline = split_text( newline, b->line[ b->end ], b->twidth );
        fprintf( stderr, "%d: %s\n", b->end, b->line[ b->end ] );

        if( b->end < 255 && newline != NULL ) b->end++;
        else return;
    }
}

// ------------------------------------------------------------------
// break width chars off of newline and store in storeit
// return pointer to new, shorter, string fragment
// ------------------------------------------------------------------
char *split_text( char *newline, char *storeit, int width )
{
    int size = 0;


    // length of string including \0 to split
    size = strlen( newline );
    fprintf( stderr, "SIZE to split: %d\n", size );

    // is the length is only 1, just store the null character and return NULL remainder string
    if( size < 2 ) {
        storeit[ 0 ] = '\0';
        fprintf( stderr, "SPLIT: size < 2\n" );
        return NULL;
    }

    // if the length is less than or equal to the copy length, do 1 copy and return NULL remainder string
    if( size <= width ) {
        strcpy( storeit, newline );
        storeit[ width + 1 ] = '\0';
        fprintf( stderr, "SPLIT: size of text to split <= width allowed per line\n" );
        return NULL;
    }

    // copy the desired length, add a null character, and inc the source pointer ahead by the width
    strncpy( storeit, newline, width );
    storeit[ width + 1 ] = '\0';
    newline += width;
    fprintf( stderr, "SPLIT: line split: width = %d\n", width );

    return newline;
}

// ------------------------------------------------------------------
// set the line # of the top of the text block display (jump scroll)
// ------------------------------------------------------------------
void tb_setposition( struct text_block *b, int x )
{
    if( !b ) return;

    if( x >= 0 && x <= 255 ) b->start = x;
}

// ------------------------------------------------------------------
// strip newline character from input string
// ------------------------------------------------------------------
void strip_newline( char *l )
{
    int i;


    for( i = 0; i < strlen( l ); i++ ) {
        if( l[ i ] == '\n' || l[ i ] == '\r' ) l[ i ] = '\0';
    }
}
